from connector.helpers import collect_methods
from connector.integration import Integration
from connector.oai.integration import Integration as OpenApiIntegration


def collect_routes(obj: object):
    """
    Collect all methods from an object and create a route for each.
    """
    from fastapi import APIRouter

    router = APIRouter()
    commands = collect_methods(obj)
    for method in commands:
        router.add_api_route(f"/{method.__name__.replace('_', '-')}", method, methods=["POST"])
    return router


def collect_integration_routes(
    integration: Integration | OpenApiIntegration,
    prefix_app_id: bool = False,
):
    """Create API endpoint for each method in integration."""
    from fastapi import APIRouter

    router = APIRouter()
    for capability_name, capability in integration.capabilities.items():
        prefix = f"/{integration.app_id}" if prefix_app_id else ""
        # replace `-` in prefix (e.g. app_id) and capability name
        route = f"{prefix}/{capability_name.value}".replace("-", "_")
        router.add_api_route(route, capability, methods=["POST"])

    return router


def runserver(router, port: int):
    import uvicorn
    from fastapi import FastAPI

    app = FastAPI()
    app.include_router(router)

    uvicorn.run(app, port=port)



import typing as t
from datetime import datetime

import pydantic

from connector.enums import (
    ActivityEventType,
    CustomAttributeCustomizedType,
    CustomAttributeType,
)

ResponseData = t.TypeVar("ResponseData", bound=pydantic.BaseModel)


class PaginationData(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid")

    token: str
    size: int | None = None


class FoundAccountData(pydantic.BaseModel):
    integration_specific_id: str
    email: t.Optional[str] = None
    given_name: t.Optional[str] = None
    family_name: t.Optional[str] = None
    username: t.Optional[str] = None
    user_status: t.Optional[str] = None
    extra_data: t.Optional[dict[str, t.Any]] = None
    custom_attributes: t.Optional[dict[str, str]] = None


class CustomAttributeSchema(pydantic.BaseModel):
    customized_type: CustomAttributeCustomizedType
    name: str
    attribute_type: CustomAttributeType


class FoundResourceData(pydantic.BaseModel):
    integration_specific_id: str
    label: str
    resource_type: str
    extra_data: t.Optional[dict[str, t.Any]] = None


class FoundEntitlementData(pydantic.BaseModel):
    integration_specific_id: str
    integration_specific_resource_id: str
    entitlement_type: str
    is_assignable: t.Optional[bool] = None
    label: t.Optional[str] = None
    extra_data: t.Optional[dict[str, t.Any]] = None


class FoundEntitlementAssociation(pydantic.BaseModel):
    integration_specific_entitlement_id: str
    account: FoundAccountData
    integration_specific_resource_id: str


class LastActivityData(pydantic.BaseModel):
    account_id: str
    event_type: ActivityEventType
    happened_at: datetime


class CapabilitySchemaData(pydantic.BaseModel):
    argument: dict[str, t.Any]
    output: dict[str, t.Any]


class EncounteredErrorResponse(pydantic.BaseModel):
    message: str
    status_code: int | None = None
    error_code: str | None = None
    raised_by: str | None = None
    raised_in: str | None = None


class Response(pydantic.BaseModel, t.Generic[ResponseData]):
    response: ResponseData
    raw_data: dict[str, t.Any] | None = None
    page: PaginationData | None = None
    error: EncounteredErrorResponse | None = None

    @classmethod
    def from_response(
        cls,
        response: ResponseData,
        raw_data: dict[str, t.Any] | None,
        page: PaginationData | None = None,
    ) -> "Response":
        return cls(
            response=response,
            raw_data=raw_data,
            page=page,
            error=None,
        )

    @classmethod
    def from_error(cls, error: EncounteredErrorResponse) -> "Response":
        return cls(
            # TODO: this should be None in error case, but for backward compat we put it here, too
            response=error,  # type: ignore[arg-type]
            raw_data=None,
            page=None,
            error=error,
        )


class AssignEntitlementResponse(pydantic.BaseModel):
    assigned: bool


class FindEntitlementAssociationsResponse(pydantic.BaseModel):
    associations: list[FoundEntitlementAssociation]


class ListAccountsResponse(pydantic.BaseModel):
    accounts: list[FoundAccountData]


class ListEntitlementsResponse(pydantic.BaseModel):
    entitlements: list[FoundEntitlementData]


class ListResourcesResponse(pydantic.BaseModel):
    resources: list[FoundResourceData]


class ListCustomAttributesSchemaResponse(pydantic.BaseModel):
    schemas: list[CustomAttributeSchema]


class UnassignEntitlementResponse(pydantic.BaseModel):
    unassigned: bool


class ValidateCredentialsResponse(pydantic.BaseModel):
    valid: bool


class CreateAccountResponse(pydantic.BaseModel):
    created: bool


class DeleteAccountResponse(pydantic.BaseModel):
    deleted: bool


class ActivateAccountResponse(pydantic.BaseModel):
    activated: bool


class DeactivateAccountResponse(pydantic.BaseModel):
    deactivated: bool


class GetLastActivityResponse(pydantic.BaseModel):
    activities: list[LastActivityData]



"""Cases for testing ``validate_credentials`` operation."""

import typing as t

import httpx
from connector.generated import (
    Error,
    ErrorCode,
    ErrorResponse,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)
from connector.utils.test import http_error_message

from {name}.integration import BASE_URL
from {name}.serializers.request import (
    {pascal}ValidateCredentials,
    {pascal}ValidateCredentialsRequest,
)
from tests.conftest import INVALID_AUTH, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    {pascal}ValidateCredentialsRequest,
    ResponseBodyMap,
    ValidateCredentialsResponse | ErrorResponse,
]


def case_validate_credentials_200() -> Case:
    """Successful request."""
    args = {pascal}ValidateCredentialsRequest(
        request={pascal}ValidateCredentials(),
        auth=VALID_AUTH,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=1": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            )
        }}
    }}
    expected_response = ValidateCredentialsResponse(
        response=ValidatedCredentials(valid=True),
        raw_data=None,
    )
    return args, response_body_map, expected_response


def case_validate_credentials_401() -> Case:
    """Unauthorized request should fail."""

    args = {pascal}ValidateCredentialsRequest(
        request={pascal}ValidateCredentials(),
        auth=INVALID_AUTH,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=1": MockedResponse(
                status_code=httpx.codes.UNAUTHORIZED,
                response_body={{}},
            )
        }}
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                f"{{BASE_URL}}/users?limit=1",
                401,
            ),
            status_code=httpx.codes.UNAUTHORIZED,
            error_code=ErrorCode.UNAUTHORIZED,
            raised_by="HTTPStatusError",
            raised_in=f"{name}.integration:validate_credentials",
        )
    )
    return args, response_body_map, expected_response



import typing as t
from enum import Enum

import pydantic

# TODO: use `RequestDataBase` which will forbid extra parameters
RequestData = t.TypeVar("RequestData", bound=pydantic.BaseModel)


class AuthModelName(str, Enum):
    OAUTH = "oauth"
    BASIC = "basic"


class AuthModel(pydantic.BaseModel):
    model: AuthModelName


class OAuthCredentials(AuthModel):
    model: t.Literal[AuthModelName.OAUTH] = AuthModelName.OAUTH
    access_token: str = pydantic.Field(title="Access Token", description="OAuth access token")
    refresh_token: str | None = pydantic.Field(
        title="Refresh Token", description="OAuth refresh token", default=None
    )
    scope: str | None = pydantic.Field(title="Scope", description="OAuth scopes", default=None)
    # Potentially URLs


class BasicCredentials(AuthModel):
    model: t.Literal[AuthModelName.BASIC] = AuthModelName.BASIC
    username: str = pydantic.Field(title="Username", description="Username")
    password: str = pydantic.Field(title="Password", description="Password")


class PaginationArgs(pydantic.BaseModel):
    token: str | None = None
    size: int | None = None


class Request(pydantic.BaseModel, t.Generic[RequestData]):
    request: RequestData
    auth: OAuthCredentials | BasicCredentials = pydantic.Field(discriminator="model")
    page: PaginationArgs | None = None
    include_raw_data: bool = False
    request_id: str | None = None

    @property
    def pagination_token(self) -> str | None:
        if self.page is None:
            return None

        return self.page.token

    @property
    def pagination_size(self) -> int | None:
        if self.page is None:
            return None

        return self.page.size

    def get_oauth(self) -> OAuthCredentials:
        if self.auth.model == AuthModelName.OAUTH:
            return self.auth
        raise ValueError(f"Invalid authentication schema: {self.auth.model}")

    def get_basic_auth(self) -> BasicCredentials:
        if self.auth.model == AuthModelName.BASIC:
            return self.auth
        raise ValueError(f"Invalid authentication schema: {self.auth.model}")

    def get_arg(self, key: str) -> t.Any:
        return getattr(self.request, key)


class ListAccountsRequest(pydantic.BaseModel):
    custom_attributes: list[str] | None = None


class ListCustomAttributesSchemaRequest(pydantic.BaseModel):
    pass


class ValidateCredentialsRequest(pydantic.BaseModel):
    pass


class ListResourcesRequest(pydantic.BaseModel):
    resource_type: str


class ListEntitlementsRequest(pydantic.BaseModel):
    resource_type: str
    resource_integration_specific_id: str


class FindEntitlementAssociationsRequest(pydantic.BaseModel):
    pass


class AssignEntitlementRequest(pydantic.BaseModel):
    account_integration_specific_id: str
    resource_integration_specific_id: str
    resource_type: str
    entitlement_integration_specific_id: str
    entitlement_type: str


class UnassignEntitlementRequest(pydantic.BaseModel):
    account_integration_specific_id: str
    resource_integration_specific_id: str
    resource_type: str
    entitlement_integration_specific_id: str
    entitlement_type: str


class CreateAccount(pydantic.BaseModel):
    email: str | None = None
    username: str | None = None
    given_name: str | None = None
    family_name: str | None = None
    user_status: str | None = None
    extra_data: dict[str, t.Any] | None = None


class CreateAccountEntitlement(pydantic.BaseModel):
    integration_specific_id: str
    integration_specific_resource_id: str
    entitlement_type: str
    is_assignable: bool | None = None
    label: str | None = None
    extra_data: dict[str, t.Any] | None = None


class CreateAccountRequest(pydantic.BaseModel):
    account: CreateAccount
    entitlements: list[CreateAccountEntitlement]


class DeleteAccountRequest(pydantic.BaseModel):
    account_id: str


class ActivateAccountRequest(pydantic.BaseModel):
    account_id: str


class DeactivateAccountRequest(pydantic.BaseModel):
    account_id: str


class GetLastActivityRequest(pydantic.BaseModel):
    account_ids: list[str]



import logging
from abc import abstractmethod
from enum import Enum
from typing import Any, Awaitable, Callable, List, Tuple, TypeAlias

from connector.serializers.lumos import EncounteredErrorResponse
from connector.serializers.request import Request, RequestData
from connector.serializers.response import Response, ResponseData

CapabilityCallable: TypeAlias = (
    Callable[[Request[RequestData]], Response[ResponseData]]
    | Callable[[Request[RequestData]], Awaitable[Response[ResponseData]]]
)

LOGGER = logging.getLogger(__name__)


class ErrorCodes(str, Enum):
    """
    Error codes for Lumos connector.
    """

    NOT_FOUND = "not_found"
    INTERNAL_ERROR = "internal_error"
    API_ERROR = "api_error"
    UNAUTHORIZED = "unauthorized"
    BAD_REQUEST = "bad_request"
    PERMISSION_DENIED = "permission_denied"
    NOT_IMPLEMENTED = "not_implemented"
    UNEXPECTED = "unexpected_error"
    UNSUPPORTED = "unsupported_operation"
    UNKNOWN_PARAM = "unknown_parameter"
    UNKNOWN_VALUE = "unknown_value"
    INVALID_VALUE = "invalid_value"


class ConnectorError(Exception):
    """
    Base exception class for Lumos connectors.
    Preferred way to raise exceptions inside the conenctors.
    `raise ConnectorError(message, error_code)`

    message: str (Custom error message)
    error_code: str | ErrorCodes (The actual error code, eg. "internal_error")
    """

    def __init__(self, message: str, error_code: str | ErrorCodes):
        self.error_code = error_code.value if isinstance(error_code, ErrorCodes) else error_code
        self.message = message


ExceptionHandlerCallable: TypeAlias = Callable[
    [
        Exception,
        (
            Callable[[Request[RequestData]], Response[ResponseData]]
            | Callable[[Request[RequestData]], Awaitable[Response[ResponseData]]]
        ),
        Response[ResponseData],
        str | ErrorCodes | None,
    ],
    Response[ResponseData],
]


class ExceptionHandler:
    """
    Abstract class for handling exceptions. You can subclass this to create your own exception handler.
    """

    def __init__(self):
        pass

    @staticmethod
    @abstractmethod
    def handle(
        e: Exception,
        original_func: Any,
        response: Response[EncounteredErrorResponse],
        error_code: str | ErrorCodes | None = None,
    ) -> Response[EncounteredErrorResponse]:
        """
        Handle an exception. (ErrorHandler signature typing)

        e: Exception (An exception that was raised)
        original_func: FunctionType (The original method that was called, eg. validate_credentials)
        response: ErrorResp (The output of the connector call)
        error_code: str | ErrorCodes | None (The actual error code, eg. "internal_error")
        """
        return response


class DefaultHandler(ExceptionHandler):
    """
    Default exception handler that handles the basic HTTPX/GQL extraction (etc.) and chains onto the global handler.
    These are operations that are always done on the raised error.
    """

    @staticmethod
    def handle(
        e: Exception,
        original_func: Any,
        response: Response[EncounteredErrorResponse],
        error_code: str | ErrorCodes | None = None,
    ) -> Response[EncounteredErrorResponse]:
        status_code: int | None = None

        # HTTPX HTTP Status code
        if hasattr(e, "response") and hasattr(e.response, "status_code"):
            status_code = e.response.status_code
        # GraphQL error code
        if hasattr(e, "code"):
            status_code = e.code

        # Populating some base info
        response.response.message = e.message if hasattr(e, "message") else str(e)
        response.response.status_code = status_code
        # TODO: add line number
        response.response.raised_in = f"{original_func.__module__}:{original_func.__name__}"
        response.response.raised_by = f"{e.__class__.__name__}"

        # ConnectorError already has an error code attached, so we need to chain
        if isinstance(e, ConnectorError):
            response.response.error_code = (
                f"{error_code}.{e.error_code}" if error_code else f"{e.error_code}"
            )
        else:
            # Otherwise, it is an unexpected error from an app_id
            response.response.error_code = (
                f"{error_code}.{ErrorCodes.UNEXPECTED}"
                if error_code
                else f"{ErrorCodes.UNEXPECTED}"
            )

        LOGGER.debug("Method %s failed", original_func.__name__, exc_info=True)
        return response


class HTTPHandler(ExceptionHandler):
    """
    Default exception handler for simple HTTP exceptions.
    If you want to handle more complicated exceptions, you can create your own instead.
    """

    @staticmethod
    def handle(
        e: Exception,
        original_func: Any,
        response: Response[EncounteredErrorResponse],
        error_code: str | ErrorCodes | None = None,
    ) -> Response[EncounteredErrorResponse]:
        if error_code:
            app_id = error_code.split(".")[0]
        else:
            app_id = "sdk"

        match response.response.status_code:
            case 400:
                response.response.error_code = f"{app_id}.{ErrorCodes.BAD_REQUEST}"
            case 401:
                response.response.error_code = f"{app_id}.{ErrorCodes.UNAUTHORIZED}"
            case 403:
                response.response.error_code = f"{app_id}.{ErrorCodes.PERMISSION_DENIED}"
            case 404:
                response.response.error_code = f"{app_id}.{ErrorCodes.NOT_FOUND}"
            case _:
                response.response.error_code = f"{app_id}.{ErrorCodes.API_ERROR}"

        return response


ErrorMap = List[Tuple[type[Exception], type[ExceptionHandler], str | ErrorCodes | None]]



"""Test cases for ``Integration.dispatch`` function."""

import json

from connector.capability import CapabilityName
from connector.integration import Integration
from connector.serializers.request import (
    BasicCredentials,
    ListAccountsRequest,
    Request,
)
from connector.serializers.response import (
    EncounteredErrorResponse,
    ListAccountsResponse,
    Response,
)

Case = tuple[
    Integration,
    CapabilityName,
    str,
    str,
]


def case_dispatch_not_implemented_handled() -> Case:
    integration = Integration(
        app_id="test",
        auth=BasicCredentials,
        exception_handlers=[],
        handle_errors=True,
    )
    capability_name = CapabilityName.LIST_ACCOUNTS
    # don't have to care about actual request data, integration should reject the call before it touches it
    request = "{}"
    expected_response = (
        Response[EncounteredErrorResponse]
        .from_error(
            error=EncounteredErrorResponse(
                message="Capability 'list-accounts' is not implemented.",
                error_code="not_implemented",
            ),
        )
        .model_dump_json()
    )
    return integration, capability_name, request, expected_response


def case_dispatch_async_success() -> Case:
    """Calling working async method should return positive response."""
    app_id = "test"
    integration = Integration(
        app_id=app_id,
        auth=BasicCredentials,
        exception_handlers=[],
        handle_errors=False,
    )

    capability_name = CapabilityName.LIST_ACCOUNTS

    @integration.register_capability(capability_name)
    async def capability(request: Request[ListAccountsRequest]) -> Response[ListAccountsResponse]:
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(
                accounts=[],
            ),
            raw_data=None,
        )

    request = json.dumps(
        {
            "request": {},
            "auth": {
                "model": "basic",
                "username": "test",
                "password": "test",
            },
        }
    )

    expected_response = Response(
        response=ListAccountsResponse(
            accounts=[],
        ),
        raw_data=None,
        page=None,
        error=None,
    ).model_dump_json()

    return integration, capability_name, request, expected_response


def case_dispatch_sync_success() -> Case:
    """Calling working sync method should return positive response."""
    app_id = "test"
    integration = Integration(
        app_id=app_id,
        auth=BasicCredentials,
        exception_handlers=[],
        handle_errors=False,
    )

    capability_name = CapabilityName.LIST_ACCOUNTS

    @integration.register_capability(capability_name)
    def capability(request: Request[ListAccountsRequest]) -> Response[ListAccountsResponse]:
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(
                accounts=[],
            ),
            raw_data=None,
        )

    request = json.dumps(
        {
            "request": {},
            "auth": {
                "model": "basic",
                "username": "test",
                "password": "test",
            },
        }
    )

    expected_response = Response(
        response=ListAccountsResponse(
            accounts=[],
        ),
        raw_data=None,
        page=None,
        error=None,
    ).model_dump_json()

    return integration, capability_name, request, expected_response



import json
from types import FunctionType

from connector.serializers.abstract import CommandTypes, EmptyModel


def generate_capability_schema(
    methods: list[FunctionType],
) -> tuple[list[str], dict[str, CommandTypes]]:
    objects = {}
    capabilities = []
    for method in methods:
        capabilities.append(name := method.__name__.lower().replace("_", "-"))

        annotations = method.__annotations__
        output = annotations.pop("return", EmptyModel)
        if len(keys := tuple(annotations.keys())) > 0:
            argument_model = annotations[keys[0]]
        else:
            argument_model = EmptyModel

        objects[name] = CommandTypes(
            argument=json.dumps(argument_model.model_json_schema()),
            output=json.dumps(output.model_json_schema()),
        )

    return capabilities, objects



# coding: utf-8

"""
    Lumos Connectors

    The Lumos Connector API is a self-describing API which allows third-party systems to easily integrate with Lumos. It provides endpoints to retrieve information about each connector, their capabilities.  # How to use this API? To get started with the Lumos Connector API, you'll 1. want to call the `POST /list-connector-app-ids` endpoint, which gives you a list of available connector app IDs, each ID corresponds to a specific integration. 2. Once you have the relevant connector app ID, use the `POST /{connector-app-id}/info` endpoint to find out what capabilities the integration supports. This will also tell you what input parameters are needed and what output you can expect for each of the API endpoints for a specific integration.  This will be integration specific. By using this info, you can correctly format your requests and handle the responses when managing resources.

    The version of the OpenAPI document: 0.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from connector.generated.models.form_data import FormData
from typing import Optional, Set
from typing_extensions import Self

class GetAuthorizationUrl(BaseModel):
    """
    GetAuthorizationUrl
    """ # noqa: E501
    client_id: StrictStr
    scopes: List[StrictStr]
    redirect_uri: StrictStr
    state: StrictStr
    form_data: Optional[Dict[str, FormData]] = None
    __properties: ClassVar[List[str]] = ["client_id", "scopes", "redirect_uri", "state", "form_data"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of GetAuthorizationUrl from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each value in form_data (dict)
        _field_dict = {}
        if self.form_data:
            for _key_form_data in self.form_data:
                if self.form_data[_key_form_data]:
                    _field_dict[_key_form_data] = self.form_data[_key_form_data].to_dict()
            _dict['form_data'] = _field_dict
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of GetAuthorizationUrl from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "client_id": obj.get("client_id"),
            "scopes": obj.get("scopes"),
            "redirect_uri": obj.get("redirect_uri"),
            "state": obj.get("state"),
            "form_data": dict(
                (_k, FormData.from_dict(_v))
                for _k, _v in obj["form_data"].items()
            )
            if obj.get("form_data") is not None
            else None
        })
        return _obj





"""Tests for ``connector.integration`` module."""

import pytest
import pytest_cases
from connector.capability import CapabilityName
from connector.integration import (
    DuplicateCapabilityError,
    Integration,
    IntegrationCapability,
)
from connector.serializers.abstract import Info
from connector.serializers.request import (
    BasicCredentials,
    Request,
    ValidateCredentialsRequest,
)
from connector.serializers.response import (
    Response,
    ResponseData,
    ValidateCredentialsResponse,
)


@pytest_cases.parametrize_with_cases(
    ["integration", "capability_name", "request_", "exception_type"],
    cases=[
        "tests.test_dispatch_raising_cases",
    ],
)
def test_dispatch_unhandled(
    integration: Integration,
    capability_name: CapabilityName,
    request_: str,
    exception_type: type[Exception],
) -> None:
    """Test if unhandled call to unregistered capability raises."""
    with pytest.raises(exception_type):
        integration.dispatch(capability_name, request_)


@pytest_cases.parametrize_with_cases(
    ["integration", "capability_name", "request_", "expected_response"],
    cases=[
        "tests.test_dispatch_returning_cases",
    ],
)
def test_dispatch(
    integration: Integration,
    capability_name: CapabilityName,
    request_: str,
    expected_response: Response[ResponseData],
) -> None:
    actual_response = integration.dispatch(capability_name, request_)
    assert actual_response == expected_response


@pytest_cases.parametrize_with_cases(
    ["integration", "expected_info"],
    cases=[
        "tests.test_info_cases",
    ],
)
def test_info(
    integration: Integration,
    expected_info: Info,
) -> None:
    actual_info = integration.info()
    assert actual_info == expected_info


@pytest_cases.parametrize_with_cases(
    ["capability_name", "integration_capabilities"],
    cases=[
        "tests.test_register_capability_cases",
    ],
)
def test_registration(
    capability_name: CapabilityName,
    integration_capabilities: dict[CapabilityName, IntegrationCapability],
) -> None:
    assert capability_name in integration_capabilities


def test_double_registration() -> None:
    integration = Integration(
        app_id="test",
        auth=BasicCredentials,
        exception_handlers=[],
    )

    async def capability(
        request: Request[ValidateCredentialsRequest],
    ) -> Response[ValidateCredentialsResponse]:
        return Response(
            response=ValidateCredentialsResponse(valid=True),
            raw_data=None,
        )

    integration.register_capability(CapabilityName.VALIDATE_CREDENTIALS)(capability)

    with pytest.raises(DuplicateCapabilityError):
        integration.register_capability(CapabilityName.VALIDATE_CREDENTIALS)(capability)



# coding: utf-8

"""
    Lumos Connectors

    The Lumos Connector API is a self-describing API which allows third-party systems to easily integrate with Lumos. It provides endpoints to retrieve information about each connector, their capabilities.  # How to use this API? To get started with the Lumos Connector API, you'll 1. want to call the `POST /list-connector-app-ids` endpoint, which gives you a list of available connector app IDs, each ID corresponds to a specific integration. 2. Once you have the relevant connector app ID, use the `POST /{connector-app-id}/info` endpoint to find out what capabilities the integration supports. This will also tell you what input parameters are needed and what output you can expect for each of the API endpoints for a specific integration.  This will be integration specific. By using this info, you can correctly format your requests and handle the responses when managing resources.

    The version of the OpenAPI document: 0.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
from inspect import getfullargspec
import json
import pprint
import re  # noqa: F401
from pydantic import BaseModel, ConfigDict, Field, StrictStr, ValidationError, field_validator
from typing import Optional
from connector.generated.models.error_response import ErrorResponse
from connector.generated.models.find_entitlement_associations_response import FindEntitlementAssociationsResponse
from typing import Union, Any, List, Set, TYPE_CHECKING, Optional, Dict
from typing_extensions import Literal, Self
from pydantic import Field

FINDENTITLEMENTASSOCIATIONS200RESPONSE_ANY_OF_SCHEMAS = ["ErrorResponse", "FindEntitlementAssociationsResponse"]

class FindEntitlementAssociations200Response(BaseModel):
    """
    FindEntitlementAssociations200Response
    """

    # data type: FindEntitlementAssociationsResponse
    anyof_schema_1_validator: Optional[FindEntitlementAssociationsResponse] = None
    # data type: ErrorResponse
    anyof_schema_2_validator: Optional[ErrorResponse] = None
    if TYPE_CHECKING:
        actual_instance: Optional[Union[ErrorResponse, FindEntitlementAssociationsResponse]] = None
    else:
        actual_instance: Any = None
    any_of_schemas: Set[str] = { "ErrorResponse", "FindEntitlementAssociationsResponse" }

    model_config = {
        "validate_assignment": True,
        "protected_namespaces": (),
    }

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError("If a position argument is used, only 1 is allowed to set `actual_instance`")
            if kwargs:
                raise ValueError("If a position argument is used, keyword arguments cannot be used.")
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator('actual_instance')
    def actual_instance_must_validate_anyof(cls, v):
        instance = FindEntitlementAssociations200Response.model_construct()
        error_messages = []
        # validate data type: FindEntitlementAssociationsResponse
        if not isinstance(v, FindEntitlementAssociationsResponse):
            error_messages.append(f"Error! Input type `{type(v)}` is not `FindEntitlementAssociationsResponse`")
        else:
            return v

        # validate data type: ErrorResponse
        if not isinstance(v, ErrorResponse):
            error_messages.append(f"Error! Input type `{type(v)}` is not `ErrorResponse`")
        else:
            return v

        if error_messages:
            # no match
            raise ValueError("No match found when setting the actual_instance in FindEntitlementAssociations200Response with anyOf schemas: ErrorResponse, FindEntitlementAssociationsResponse. Details: " + ", ".join(error_messages))
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        # anyof_schema_1_validator: Optional[FindEntitlementAssociationsResponse] = None
        try:
            instance.actual_instance = FindEntitlementAssociationsResponse.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
             error_messages.append(str(e))
        # anyof_schema_2_validator: Optional[ErrorResponse] = None
        try:
            instance.actual_instance = ErrorResponse.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
             error_messages.append(str(e))

        if error_messages:
            # no match
            raise ValueError("No match found when deserializing the JSON string into FindEntitlementAssociations200Response with anyOf schemas: ErrorResponse, FindEntitlementAssociationsResponse. Details: " + ", ".join(error_messages))
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(self.actual_instance.to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(self) -> Optional[Union[Dict[str, Any], ErrorResponse, FindEntitlementAssociationsResponse]]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        else:
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())





"""Test cases for ``Integration.info`` function."""

import typing as t

from connector.capability import CapabilityName
from connector.integration import Integration, IntegrationCapability
from connector.serializers.request import BasicCredentials, ListAccountsRequest, Request
from connector.serializers.response import ListAccountsResponse, Response

Case: t.TypeAlias = tuple[
    CapabilityName,
    dict[CapabilityName, IntegrationCapability],
]


def case_register_capability_success() -> Case:
    app_id = "test"
    integration = Integration(
        app_id=app_id,
        auth=BasicCredentials,
        exception_handlers=[],
        handle_errors=True,
    )

    capability_name = CapabilityName.LIST_ACCOUNTS

    @integration.register_capability(capability_name)
    async def capability(request: Request[ListAccountsRequest]) -> Response[ListAccountsResponse]:
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(
                accounts=[],
            ),
            raw_data=None,
        )

    return capability_name, integration.capabilities



import inspect
import json
import subprocess
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import cast

from connector.capability import CapabilityName
from connector.generated import CapabilityName as OpenApiCapabilityName
from connector.helpers import collect_methods, get_pydantic_model
from connector.integration import Integration
from connector.oai.integration import Integration as OpenApiIntegration

__all__ = ("collect_commands", "run_command")


# Hacking commands
# ----------------


def _prep_hacking_command(args: Namespace):
    data = vars(args)
    data.pop("command")
    data.pop("func")
    return data


def unasync(
    sync_commands: object, directory: str = "", check: bool = False, quiet: bool = False
) -> None:
    from connector.make_sync import make_sync

    if not directory:
        _dir = Path(inspect.getfile(sync_commands.__class__)).parent.parent / "async_"
    else:
        _dir = Path(directory)

    success = make_sync(Path(_dir), check=check, quiet=quiet)
    if check:
        if not quiet:
            print("Unasync check complete")
        exit(bool(not success))
    else:
        if not quiet:
            print("Unasync complete")


def http_command_server(async_commands: object, port: int = 8000):
    from connector.http_server import collect_routes, runserver

    router = collect_routes(async_commands)
    try:
        runserver(router, port)
    except KeyboardInterrupt:
        pass


def http_integration_server(integration: Integration | OpenApiIntegration, port: int = 8000):
    from connector.http_server import collect_integration_routes, runserver

    router = collect_integration_routes(integration)
    try:
        runserver(router, port)
    except KeyboardInterrupt:
        pass


def build_executable(path: str) -> None:
    try:
        subprocess.run(["pyinstaller", "--version"], check=True)
    except FileNotFoundError:
        print("PyInstaller not found in PATH. Please pip install pyinstaller")
        return

    command = [
        "pyinstaller",
        path,
        "--noconsole",
        "--onefile",
        "--clean",
        "--paths=projects/libs/python",
    ]
    if __file__ not in "site-packages":
        command.append("--paths=projects/libs/python")
    subprocess.run(command)


def run_test():
    subprocess.run(["pytest", "tests/"], check=True)


def create_command_hacking_parser(
    sync_commands: object, async_commands: object, parser: ArgumentParser
) -> None:
    subparsers = parser.add_subparsers(dest="command")

    unasync_parser = subparsers.add_parser("unasync")
    unasync_parser.add_argument(
        "--directory",
        "-d",
        type=str,
        help="The directory to unasync. Defaults to the async_ directory in the connector.",
    )
    unasync_parser.add_argument(
        "--check",
        "-c",
        action="store_true",
        help="Check if the async functions can be converted to sync.",
    )
    unasync_parser.add_argument(
        "--quiet", "-q", action="store_true", help="Do not print anything to the console."
    )
    unasync_parser.set_defaults(
        func=lambda args: unasync(sync_commands, **_prep_hacking_command(args))
    )

    http_server_parser = subparsers.add_parser("http-server")
    http_server_parser.add_argument(
        "--port", "-p", type=int, default=8000, help="The port to run the server on."
    )
    http_server_parser.set_defaults(
        func=lambda args: http_command_server(async_commands, **_prep_hacking_command(args))
    )

    build_executable_parser = subparsers.add_parser(
        "build-executable",
        help=(
            "Create a single file executable with PyInstaller. Provide the path to your library's"
            " main.py file."
        ),
    )
    build_executable_parser.add_argument("path", type=str, help="The path to the main.py file.")
    build_executable_parser.set_defaults(
        func=lambda args: build_executable(**_prep_hacking_command(args))
    )

    test_parser = subparsers.add_parser("test")
    test_parser.set_defaults(func=lambda args: run_test())

    return None


def create_integration_hacking_parser(
    integration: Integration | OpenApiIntegration, parser: ArgumentParser
) -> None:
    subparsers = parser.add_subparsers(dest="command")

    http_server_parser = subparsers.add_parser("http-server")
    http_server_parser.add_argument(
        "--port", "-p", type=int, default=8000, help="The port to run the server on."
    )
    http_server_parser.set_defaults(
        func=lambda args: http_integration_server(integration, **_prep_hacking_command(args))
    )

    build_executable_parser = subparsers.add_parser(
        "build-executable",
        help=(
            "Create a single file executable with PyInstaller. Provide the path to your library's"
            " main.py file."
        ),
    )
    build_executable_parser.add_argument("path", type=str, help="The path to the main.py file.")
    build_executable_parser.set_defaults(
        func=lambda args: build_executable(**_prep_hacking_command(args))
    )

    test_parser = subparsers.add_parser("test")
    test_parser.set_defaults(func=lambda args: run_test())

    return None


# Actual Commands
# ---------------


def _print_pydantic(model):
    # Pydantic v2
    if hasattr(model, "model_dump_json"):
        print(model.model_dump_json())
    # Pydantic v1
    elif hasattr(model, "json"):
        print(model.json())
    elif type(model) in (dict, list):
        print(json.dumps(model))
    else:
        print(model)


def _print_decorator(func):
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if isinstance(result, list):
            for item in result:
                _print_pydantic(item)
        else:
            _print_pydantic(result)
        return result

    return wrapper


def command_executor(sync_commands: object, args: Namespace):
    """Executes a command from the CLI."""
    method = getattr(sync_commands, cast(str, args.command).replace("-", "_"))
    try:
        model_cls = get_pydantic_model(method.__annotations__)
    except ValueError:
        model_cls = None

    if model_cls:
        try:
            model = model_cls.model_validate_json(args.json)
        except AttributeError:
            model = model_cls.parse_raw(args.json)
        output = method(model)
    else:
        output = method()
    _print_pydantic(output)


def capability_executor(integration: Integration | OpenApiIntegration, args: Namespace):
    """Executes a command from the CLI."""
    if isinstance(integration, OpenApiIntegration):
        output = integration.dispatch(OpenApiCapabilityName(args.command), args.json)
    else:
        output = integration.dispatch(CapabilityName(args.command), args.json)
    print(output)


def collect_commands(
    sync_commands: object, async_commands: object, no_print: bool = False
) -> ArgumentParser:
    """
    Collect all methods from an object and create a CLI command for each.
    The object must be instantiated.
    """
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")

    commands = collect_methods(sync_commands)
    for method in commands:
        subparser = subparsers.add_parser(
            method.__name__.replace("_", "-"), description=method.__doc__
        )

        try:
            get_pydantic_model(method.__annotations__)
        except ValueError:
            pass
        else:
            subparser.add_argument("--json", type=str, help="JSON input", required=True)

        subparser.set_defaults(func=lambda args: command_executor(sync_commands, args))

    hacking_subparser = subparsers.add_parser("hacking")
    create_command_hacking_parser(sync_commands, async_commands, hacking_subparser)

    return parser


def collect_capabilities(
    integration: Integration | OpenApiIntegration, no_print: bool = False
) -> ArgumentParser:
    """
    Collect all methods from an Integration class and create a CLI
    command for each.
    """
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")

    subparser = subparsers.add_parser("info", description=integration.info.__doc__)
    subparser.set_defaults(func=lambda args: command_executor(integration, args))

    for capability_name, capability in integration.capabilities.items():
        subparser = subparsers.add_parser(capability_name.value, description=capability.__doc__)

        try:
            get_pydantic_model(capability.__annotations__)
        except ValueError:
            pass
        else:
            subparser.add_argument("--json", type=str, help="JSON input", required=True)

        subparser.set_defaults(func=lambda args: capability_executor(integration, args))

    hacking_subparser = subparsers.add_parser("hacking")
    create_integration_hacking_parser(integration, hacking_subparser)

    return parser


def run_command(sync_commands: object, async_commands: object, no_print: bool = False) -> None:
    """Run a command from the CLI."""
    parser = collect_commands(sync_commands, async_commands, no_print)
    args = parser.parse_args()
    args.func(args)


def run_integration(
    integration: Integration | OpenApiIntegration,
    no_print: bool = False,
) -> None:
    """Run a command from the CLI, integratin version."""
    parser = collect_capabilities(integration, no_print)
    args = parser.parse_args()
    args.func(args)



from connector.generated import (
    ActivateAccount,
    ActivateAccountRequest,
    AssignEntitlement,
    AssignEntitlementRequest,
    CreateAccount,
    CreateAccountEntitlement,
    CreateAccountRequest,
    DeactivateAccount,
    DeactivateAccountRequest,
    DeleteAccount,
    DeleteAccountRequest,
    FindEntitlementAssociations,
    FindEntitlementAssociationsRequest,
    GetLastActivity,
    GetLastActivityRequest,
    ListAccounts,
    ListAccountsRequest,
    ListCustomAttributesSchema,
    ListCustomAttributesSchemaRequest,
    ListEntitlements,
    ListEntitlementsRequest,
    ListResources,
    ListResourcesRequest,
    UnassignEntitlement,
    UnassignEntitlementRequest,
    ValidateCredentials,
    ValidateCredentialsRequest,
)

class {pascal}ActivateAccount(ActivateAccount):
    pass

class {pascal}ActivateAccountRequest(ActivateAccountRequest):
    request: {pascal}ActivateAccount

class {pascal}AssignEntitlement(AssignEntitlement):
    pass

class {pascal}AssignEntitlementRequest(AssignEntitlementRequest):
    request: {pascal}AssignEntitlement

class {pascal}CreateAccount(CreateAccount):
    pass

class {pascal}CreateAccountRequest(CreateAccountRequest):
    request: {pascal}CreateAccount

class {pascal}DeactivateAccount(DeactivateAccount):
    pass

class {pascal}DeactivateAccountRequest(DeactivateAccountRequest):
    request: {pascal}DeactivateAccount

class {pascal}DeleteAccount(DeleteAccount):
    pass

class {pascal}DeleteAccountRequest(DeleteAccountRequest):
    request: {pascal}DeleteAccount

class {pascal}FindEntitlementAssociations(FindEntitlementAssociations):
    pass

class {pascal}FindEntitlementAssociationsRequest(FindEntitlementAssociationsRequest):
    request: {pascal}FindEntitlementAssociations

class {pascal}GetLastActivity(GetLastActivity):
    pass

class {pascal}GetLastActivityRequest(GetLastActivityRequest):
    request: {pascal}GetLastActivity

class {pascal}ListAccounts(ListAccounts):
    pass

class {pascal}ListAccountsRequest(ListAccountsRequest):
    request: {pascal}ListAccounts

class {pascal}ListCustomAttributesSchema(ListCustomAttributesSchema):
    pass

class {pascal}ListCustomAttributesSchemaRequest(ListCustomAttributesSchemaRequest):
    request: {pascal}ListCustomAttributesSchema

class {pascal}ListEntitlements(ListEntitlements):
    pass

class {pascal}ListEntitlementsRequest(ListEntitlementsRequest):
    request: {pascal}ListEntitlements

class {pascal}ListResources(ListResources):
    pass

class {pascal}ListResourcesRequest(ListResourcesRequest):
    request: {pascal}ListResources

class {pascal}UnassignEntitlement(UnassignEntitlement):
    pass

class {pascal}UnassignEntitlementRequest(UnassignEntitlementRequest):
    request: {pascal}UnassignEntitlement

class {pascal}ValidateCredentials(ValidateCredentials):
    pass

class {pascal}ValidateCredentialsRequest(ValidateCredentialsRequest):
    request: {pascal}ValidateCredentials



import pytest
import pytest_cases
from connector.oai.capability import CapabilityName, get_oauth

from connector.generated import (
    CreateAccountResponse,
    Error,
    ErrorCode,
    ErrorResponse,
)
from {name}.integration import integration
from {name}.serializers.request import (
    {pascal}CreateAccountRequest,
)

from tests.type_definitions import ClientContextManager, ResponseBodyMap


@pytest.mark.skip(reason="Function not implemented yet, remove after implementation of tested function.")
@pytest_cases.parametrize_with_cases(
    ["args", "response_body_map", "expected_response"],
    cases=["tests.test_account_provisioning.test_create_account_cases"],
)
def test_create_account(
    httpx_async_client: ClientContextManager,
    args: {pascal}CreateAccountRequest,
    response_body_map: ResponseBodyMap,
    expected_response: CreateAccountResponse | ErrorResponse,
) -> None:
    """Test ``create_account`` operation."""
    with httpx_async_client(get_oauth(args).access_token, response_body_map):
        response = integration.dispatch(CapabilityName.CREATE_ACCOUNT, args.model_dump_json())

    assert response == expected_response.model_dump_json()



"""Utilities for describing capabilities.

Each known capability is assigned a base class for request and response.
The actual request and response types in a integration implementation
can either use the base classes directly or create subclasses, however,
those bases are enforced to be used.
"""

import enum
import inspect
import json
import typing as t

from connector.generated import (
    AuthorizationUrl,
    GetAuthorizationUrl,
    HandleAuthorizationCallback,
    OauthCredentials as GenOauthCredentials,
    RefreshAccessToken,
)
from connector.serializers.abstract import CommandTypes
from connector.serializers.request import (
    ActivateAccountRequest,
    AssignEntitlementRequest,
    CreateAccountRequest,
    DeactivateAccountRequest,
    DeleteAccountRequest,
    FindEntitlementAssociationsRequest,
    GetLastActivityRequest,
    ListAccountsRequest,
    ListCustomAttributesSchemaRequest,
    ListEntitlementsRequest,
    ListResourcesRequest,
    Request,
    RequestData,
    UnassignEntitlementRequest,
    ValidateCredentialsRequest,
)
from connector.serializers.response import (
    ActivateAccountResponse,
    AssignEntitlementResponse,
    CreateAccountResponse,
    DeactivateAccountResponse,
    DeleteAccountResponse,
    FindEntitlementAssociationsResponse,
    GetLastActivityResponse,
    ListAccountsResponse,
    ListCustomAttributesSchemaResponse,
    ListEntitlementsResponse,
    ListResourcesResponse,
    Response,
    ResponseData,
    UnassignEntitlementResponse,
    ValidateCredentialsResponse,
)


class CapabilityName(str, enum.Enum):
    """Enumeration of known capabilities."""

    ACTIVATE_ACCOUNT = "activate_account"
    ASSIGN_ENTITLEMENT = "assign-entitlement"
    CREATE_ACCOUNT = "create_account"
    DEACTIVATE_ACCOUNT = "deactivate_account"
    DELETE_ACCOUNT = "delete_account"
    FIND_ENTITLEMENT_ASSOCIATIONS = "find-entitlement-associations"
    GET_LAST_ACTIVITY = "get-last-activity"
    GET_AUTHORIZATION_URL = "get_authorization_url"
    LIST_ACCOUNTS = "list-accounts"
    HANDLE_AUTHORIZATION_CALLBACK = "handle_authorization_callback"
    LIST_CUSTOM_ATTRIBUTES_SCHEMA = "list-custom-attributes-schema"
    LIST_ENTITLEMENTS = "list-entitlements"
    LIST_RESOURCES = "list-resources"
    REFRESH_ACCESS_TOKEN = "refresh_access_token"
    UNASSIGN_ENTITLEMENT = "unassign-entitlement"
    VALIDATE_CREDENTIALS = "validate-credentials"


CAPABILITY_REQUEST_BASES = {
    CapabilityName.ACTIVATE_ACCOUNT: ActivateAccountRequest,
    CapabilityName.ASSIGN_ENTITLEMENT: AssignEntitlementRequest,
    CapabilityName.CREATE_ACCOUNT: CreateAccountRequest,
    CapabilityName.DEACTIVATE_ACCOUNT: DeactivateAccountRequest,
    CapabilityName.DELETE_ACCOUNT: DeleteAccountRequest,
    CapabilityName.FIND_ENTITLEMENT_ASSOCIATIONS: FindEntitlementAssociationsRequest,
    CapabilityName.GET_AUTHORIZATION_URL: GetAuthorizationUrl,
    CapabilityName.GET_LAST_ACTIVITY: GetLastActivityRequest,
    CapabilityName.HANDLE_AUTHORIZATION_CALLBACK: HandleAuthorizationCallback,
    CapabilityName.LIST_ACCOUNTS: ListAccountsRequest,
    CapabilityName.LIST_CUSTOM_ATTRIBUTES_SCHEMA: ListCustomAttributesSchemaRequest,
    CapabilityName.LIST_ENTITLEMENTS: ListEntitlementsRequest,
    CapabilityName.LIST_RESOURCES: ListResourcesRequest,
    CapabilityName.REFRESH_ACCESS_TOKEN: RefreshAccessToken,
    CapabilityName.UNASSIGN_ENTITLEMENT: UnassignEntitlementRequest,
    CapabilityName.VALIDATE_CREDENTIALS: ValidateCredentialsRequest,
}
CAPABILITY_RESPONSE_BASES = {
    CapabilityName.ACTIVATE_ACCOUNT: ActivateAccountResponse,
    CapabilityName.ASSIGN_ENTITLEMENT: AssignEntitlementResponse,
    CapabilityName.CREATE_ACCOUNT: CreateAccountResponse,
    CapabilityName.DEACTIVATE_ACCOUNT: DeactivateAccountResponse,
    CapabilityName.DELETE_ACCOUNT: DeleteAccountResponse,
    CapabilityName.FIND_ENTITLEMENT_ASSOCIATIONS: FindEntitlementAssociationsResponse,
    CapabilityName.GET_AUTHORIZATION_URL: AuthorizationUrl,
    CapabilityName.GET_LAST_ACTIVITY: GetLastActivityResponse,
    CapabilityName.HANDLE_AUTHORIZATION_CALLBACK: GenOauthCredentials,
    CapabilityName.LIST_ACCOUNTS: ListAccountsResponse,
    CapabilityName.LIST_CUSTOM_ATTRIBUTES_SCHEMA: ListCustomAttributesSchemaResponse,
    CapabilityName.LIST_ENTITLEMENTS: ListEntitlementsResponse,
    CapabilityName.LIST_RESOURCES: ListResourcesResponse,
    CapabilityName.REFRESH_ACCESS_TOKEN: GenOauthCredentials,
    CapabilityName.UNASSIGN_ENTITLEMENT: UnassignEntitlementResponse,
    CapabilityName.VALIDATE_CREDENTIALS: ValidateCredentialsResponse,
}


def generate_capability_schema(
    capability_name: str,
    capability: (
        t.Callable[[Request[RequestData]], Response[ResponseData]]
        | t.Callable[[Request[RequestData]], t.Awaitable[Response[ResponseData]]]
    ),
) -> CommandTypes:
    request_annotation, response_annotation = get_capability_annotations(capability)
    return CommandTypes(
        argument=json.dumps(request_annotation.model_json_schema()),
        output=json.dumps(response_annotation.model_json_schema()),
    )


def get_capability_annotations(
    capability: (
        t.Callable[[Request[RequestData]], Response[ResponseData]]
        | t.Callable[[Request[RequestData]], t.Awaitable[Response[ResponseData]]]
    ),
) -> tuple[type[Request[RequestData]], type[Response[ResponseData]]]:
    """Extract argument and return type annotations."""
    annotations = inspect.get_annotations(capability)
    try:
        response_annotation = annotations["return"]
        request_annotation_name = (set(annotations.keys()) - {"return"}).pop()
    except KeyError:
        raise TypeError(
            f"The capability function {capability.__name__} must have both request and return annotations."
        ) from None

    request_annotation = annotations[request_annotation_name]
    return request_annotation, response_annotation


def validate_capability(
    name: CapabilityName,
    capability: (
        t.Callable[[Request[RequestData]], Response[ResponseData]]
        | t.Callable[[Request[RequestData]], t.Awaitable[Response[ResponseData]]]
    ),
) -> None:
    """Make sure copability implementation is valid.

    Capability is marked as valid when:
        * is fully annotated, i.e., both argument and return value are
        type-hinted
        * type of accepted argument matches the expected one, i.e., is
        exactly the same class or a subclass
        * type of returned value matches the expected one, same
        mechanism as for argument
    """
    request_annotation, response_annotation = get_capability_annotations(capability)
    expected_response_model = CAPABILITY_RESPONSE_BASES[name]
    actual_response_model = t.cast(
        type[ResponseData],
        response_annotation.model_fields["response"].annotation,
    )
    if actual_response_model != expected_response_model:
        raise TypeError(
            f"The function {capability.__name__} for capability {name} must return {expected_response_model.__name__}. "
            f"Actual response model: {actual_response_model.__name__}"
        ) from None

    expected_request_model = CAPABILITY_REQUEST_BASES[name]
    actual_request_model = t.cast(
        type[RequestData],
        request_annotation.model_fields["request"].annotation,
    )
    if not issubclass(actual_request_model, expected_request_model):
        raise TypeError(
            f"The function {capability.__name__} for capability {name} must accept {expected_request_model.__name__} "
            f"or its subclass. Actual request model: {actual_request_model.__name__}"
        ) from None



# coding: utf-8

"""
    Lumos Connectors

    The Lumos Connector API is a self-describing API which allows third-party systems to easily integrate with Lumos. It provides endpoints to retrieve information about each connector, their capabilities.  # How to use this API? To get started with the Lumos Connector API, you'll 1. want to call the `POST /list-connector-app-ids` endpoint, which gives you a list of available connector app IDs, each ID corresponds to a specific integration. 2. Once you have the relevant connector app ID, use the `POST /{connector-app-id}/info` endpoint to find out what capabilities the integration supports. This will also tell you what input parameters are needed and what output you can expect for each of the API endpoints for a specific integration.  This will be integration specific. By using this info, you can correctly format your requests and handle the responses when managing resources.

    The version of the OpenAPI document: 0.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class CustomAttributeCustomizedType(str, Enum):
    """
    CustomAttributeCustomizedType
    """

    """
    allowed enum values
    """
    ACCOUNT = 'account'

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of CustomAttributeCustomizedType from a JSON string"""
        return cls(json.loads(json_str))





"""Utility class to manage connector implementation.

:py:class:`Integration` provides a single point to register Integration
capabilities.

By instantiating the :py:class:`Integration` you simply create a basic
integration without any real implementation. To actually implement any
capability, you have to define (async) function outside the class and
register them to the integration instance by decorating the
implementation with ``@integration.register_capability(name)``.

Capability function has to:
    * accept only one argument
    * return scalar response

The :py:class:`Integration` is as much type-hinted as possible and also
does several checks to ensure that the implementation "is correct".
Incorrect implementation should raise an error during application start
(fail fast).

What is checked (at application start):
    * capability name is known (defined in ``CapabilityName`` enum)
    * the typesu of accepted argument and returned value matches the
    capability interface
"""

import asyncio
import inspect
import json
import typing as t
from dataclasses import dataclass

from connector.async_.exception_handler import exception_handler as async_exception_handler
from connector.capability import (
    CapabilityName,
    generate_capability_schema,
    get_capability_annotations,
    validate_capability,
)
from connector.errors import ErrorCodes, ErrorMap
from connector.serializers.abstract import AppCategory, Info
from connector.serializers.request import (
    BasicCredentials,
    OAuthCredentials,
    Request,
    RequestData,
)
from connector.serializers.response import EncounteredErrorResponse, Response, ResponseData
from connector.sync_.exception_handler import exception_handler as sync_exception_handler

CapabilityCallable: t.TypeAlias = (
    t.Callable[[Request[RequestData]], Response[ResponseData]]
    | t.Callable[[Request[RequestData]], t.Awaitable[Response[ResponseData]]]
)
IntegrationCapability = (
    t.Callable[[Request[t.Any]], Response[t.Any]]
    | t.Callable[[Request[t.Any]], t.Awaitable[Response[t.Any]]]
)
AuthSetting: t.TypeAlias = t.Union[
    t.Type[OAuthCredentials],
    t.Type[BasicCredentials],
    None,  # TODO: Remove once all connectors are migrated to new capability registration
]


class IntegrationError(Exception):
    """Base class for exceptions raised by Integration."""


class DuplicateCapabilityError(IntegrationError):
    """Raised when registering the same capability repeatedly."""

    def __init__(self, capability_name: CapabilityName) -> None:
        super().__init__(f"{capability_name} already registered")


class InvalidAppIdError(IntegrationError):
    """Raised when app_id is not valid.

    Most probably, empty or containing only whitespaces.
    """


@dataclass
class DescriptionData:
    logo_url: str | None = None
    user_friendly_name: str | None = None
    description: str | None = None
    categories: list[AppCategory] | None = None


class Integration:
    app_id: str

    def __init__(
        self,
        app_id: str,
        exception_handlers: ErrorMap,
        auth: AuthSetting,
        handle_errors: bool = True,
        description_data: DescriptionData | None = None,
    ):
        self.app_id = app_id.strip()
        self.description_data = description_data or DescriptionData()
        self.auth = auth
        self.exception_handlers = exception_handlers
        self.handle_errors = handle_errors

        if len(self.app_id) == 0:
            raise InvalidAppIdError

        self.capabilities: dict[CapabilityName, IntegrationCapability] = {}

    def register_capability(
        self,
        name: CapabilityName,
    ) -> t.Callable[[CapabilityCallable], CapabilityCallable]:
        """Add implementation of specified capability.

        This function is expected to be used as a decorator for a
        capability implementation.

        Raises
        ------
        DuplicateCapabilityError:
            When capability is registered more that once.
        """
        if name in self.capabilities:
            raise DuplicateCapabilityError(name)

        def decorator(
            func: CapabilityCallable,
        ) -> CapabilityCallable:
            validate_capability(name, func)
            self.capabilities[name] = func
            return func

        return decorator

    def dispatch(self, name: CapabilityName, request_string: str) -> str:
        """Call implemented capability, returning the result.

        Raises
        ------
        NotImplementedError:
            When capability is not implemented (or registered)
        """
        try:
            capability = self.capabilities[name]
        except KeyError:
            if self.handle_errors:
                return (
                    Response[EncounteredErrorResponse]
                    .from_error(
                        error=EncounteredErrorResponse(
                            message=f"Capability '{name.value}' is not implemented.",
                            error_code=ErrorCodes.NOT_IMPLEMENTED.value,
                        ),
                    )
                    .model_dump_json()
                )

            raise NotImplementedError from None

        request_annotation, _ = get_capability_annotations(capability)
        request = request_annotation(**json.loads(request_string))

        if inspect.iscoroutinefunction(capability):
            response = asyncio.run(
                async_exception_handler(self.exception_handlers, self.app_id)(capability)(request)
                if self.handle_errors
                else capability(request)
            )
        else:
            response = (
                sync_exception_handler(self.exception_handlers, self.app_id)(capability)(request)
                if self.handle_errors
                else capability(request)
            )

        return response.model_dump_json()

    def info(self) -> Info:
        """Provide information about implemented capabilities.

        Json schema describing implemented capabilities and their
        interface is returned. The authentication schema is also
        included.
        """
        capability_names = sorted(capability_name.value for capability_name in self.capabilities)
        return Info(
            app_id=self.app_id,
            capabilities=capability_names,
            # TODO: Remove None check once all connectors are migrated to new capability registration
            authentication_schema=self.auth.model_json_schema() if self.auth else {},
            capability_schema={
                capability_name: generate_capability_schema(
                    capability_name, self.capabilities[capability_name]
                )
                for capability_name in sorted(self.capabilities)
            },
            logo_url=self.description_data.logo_url,
            user_friendly_name=self.description_data.user_friendly_name,
            description=self.description_data.description,
            categories=self.description_data.categories,
        )



"""Test cases for ``get_capability_annotations`` function."""

import typing as t

import pytest_cases
from connector.serializers.request import Request, RequestData, ValidateCredentialsRequest
from connector.serializers.response import Response, ResponseData, ValidateCredentialsResponse

Case: t.TypeAlias = tuple[
    t.Callable[[Request[RequestData]], Response[ResponseData]],
    tuple[type[Request[RequestData]], type[Response[ResponseData]]] | None,
]


@pytest_cases.case(tags=("correct",))
def case_correct_capability() -> Case:
    def capability(
        args: Request[ValidateCredentialsRequest],
    ) -> Response[ValidateCredentialsResponse]:
        return Response[ValidateCredentialsResponse](
            response=ValidateCredentialsResponse(valid=True),
            raw_data=None,
        )

    expected_annotations = (
        Request[ValidateCredentialsRequest],
        Response[ValidateCredentialsResponse],
    )
    return capability, expected_annotations


@pytest_cases.case(tags=("missing_annotation",))
def case_missing_argument_annotation() -> Case:
    def capability(args) -> Response[ValidateCredentialsResponse]:
        return Response[ValidateCredentialsResponse](
            response=ValidateCredentialsResponse(valid=True),
            raw_data=None,
        )

    return capability, None


@pytest_cases.case(tags=("missing_annotation",))
def case_missing_return_annotation() -> Case:
    def capability(args: Request[ValidateCredentialsRequest]):
        return Response[ValidateCredentialsResponse](
            response=ValidateCredentialsResponse(valid=True),
            raw_data=None,
        )

    return capability, None


@pytest_cases.case(tags=("missing_annotation",))
def case_missing_annotations() -> Case:
    def capability(args):
        return Response[ValidateCredentialsResponse](
            response=ValidateCredentialsResponse(valid=True),
            raw_data=None,
        )

    return capability, None



import argparse
from pathlib import Path


def main():
    parser = argparse.ArgumentParser(description="(Lumos) Connectors CLI")
    subparsers = parser.add_subparsers(dest="command")

    scaffold_parser = subparsers.add_parser("scaffold", help="Create a new connector")
    scaffold_parser.add_argument("name", help="Name of the new connector")
    scaffold_parser.add_argument(
        "directory", type=Path, help="Directory to create the connector in"
    )

    hacking_parser = subparsers.add_parser(
        "hacking", help="Run commands for developing the `connector` library"
    )
    hacking_subparsers = hacking_parser.add_subparsers(dest="hacking_command")
    hacking_subparsers.add_parser("unasync", help="Unasync async code")

    args = parser.parse_args()

    if args.command == "scaffold":
        from connector.scaffold.create import scaffold

        scaffold(args)
    elif args.command == "hacking":
        if args.hacking_command == "unasync":
            from connector.make_sync import make_sync

            make_sync(Path(__file__).parent / "async_")
            make_sync(Path(__file__).parent / "scaffold" / "templates" / "connector" / "async_")


if __name__ == "__main__":
    main()



import enum


class ActivityEventType(str, enum.Enum):
    LAST_LOGIN = "LastLogin"
    LAST_ACTIVITY = "LastActivity"


class CustomAttributeType(str, enum.Enum):
    STRING = "STRING"
    USER = "USER"


class CustomAttributeCustomizedType(str, enum.Enum):
    ACCOUNT = "ACCOUNT"
    ENTITLEMENMT = "ENTITLEMENT"
    RESOURCE = "RESOURCE"



"""Test cases for ``validate_capability`` function."""

import typing as t

import pytest_cases
from connector.capability import CapabilityName
from connector.serializers.request import (
    ListAccountsRequest,
    ListResourcesRequest,
    Request,
    RequestData,
)
from connector.serializers.response import (
    ListAccountsResponse,
    Response,
    ResponseData,
)

Case: t.TypeAlias = tuple[
    CapabilityName,
    t.Callable[[Request[RequestData]], Response[ResponseData]],
]


class CustomListAccountsRequest(ListAccountsRequest):
    """Correct custom request for list-accounts capability."""

    extra: str


class CustomListAccountsRequestFromBadBase(ListResourcesRequest):
    """Incorrect base is used for custom request schema."""

    extra: str


class CustomListAccountsResponse(ListAccountsResponse):
    """Subclassing response type is always bad."""

    extra_resp: str


@pytest_cases.case(tags=("valid",))
def case_valid_capability_base_annotation() -> Case:
    def capability(request: Request[ListAccountsRequest]) -> Response[ListAccountsResponse]:
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(accounts=[]),
            raw_data=None,
        )

    capability_name = CapabilityName.LIST_ACCOUNTS
    return capability_name, capability


@pytest_cases.case(tags=("valid",))
def case_valid_capability_custom_request() -> Case:
    """Using subclass of matching request type is valid.

    Using subclass of ListAccountsRequest for list-accounts capability
    is absolutely fine.
    """

    def capability(request: Request[CustomListAccountsRequest]) -> Response[ListAccountsResponse]:
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(accounts=[]),
            raw_data=None,
        )

    capability_name = CapabilityName.LIST_ACCOUNTS
    return capability_name, capability


@pytest_cases.case(tags=("invalid",))
def case_invalid_capability_custom_response() -> Case:
    """Using subclass of SDK defined response is not correct.

    This would change the output of the method, making it super hard to
    use.
    """

    def capability(request: Request[ListAccountsRequest]) -> Response[CustomListAccountsResponse]:
        return Response[CustomListAccountsResponse](
            response=CustomListAccountsResponse(accounts=[], extra_resp="bad"),
            raw_data=None,
        )

    capability_name = CapabilityName.LIST_ACCOUNTS
    return capability_name, capability


@pytest_cases.case(tags=("invalid",))
def case_invalid_capability_bad_request_model() -> Case:
    """Using mismatching request type for capability is invalid.

    Using ListResourcesRequest for list-accounts method is obviously
    invalid.
    """

    def capability(request: Request[ListResourcesRequest]) -> Response[ListAccountsResponse]:
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(accounts=[]),
            raw_data=None,
        )

    capability_name = CapabilityName.LIST_ACCOUNTS
    return capability_name, capability


@pytest_cases.case(tags=("invalid",))
def case_invalid_capability_bad_request_base_model() -> Case:
    """Using mismatching base for request type is invalid.

    Using subslass of ListResourcesRequest for list-accounts method is
    obviously invalid.
    """

    def capability(
        request: Request[CustomListAccountsRequestFromBadBase],
    ) -> Response[ListAccountsResponse]:
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(accounts=[]),
            raw_data=None,
        )

    capability_name = CapabilityName.LIST_ACCOUNTS
    return capability_name, capability


@pytest_cases.case(tags=("invalid",))
def case_invalid_capability_bad_request_base() -> Case:
    """Using class unrelated to request type is invalid.

    Using classes unrelated to ``RequestData`` is obviously invalid.
    """

    def capability(request: Request[int]) -> Response[ListAccountsResponse]:  # type: ignore[type-var]
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(accounts=[]),
            raw_data=None,
        )

    capability_name = CapabilityName.LIST_ACCOUNTS
    return capability_name, capability



"""Test cases for ``Integration.info`` function."""

import json
import typing as t

from connector.capability import CapabilityName
from connector.integration import Integration
from connector.serializers.abstract import CommandTypes, Info
from connector.serializers.request import BasicCredentials, ListAccountsRequest, Request
from connector.serializers.response import ListAccountsResponse, Response

Case: t.TypeAlias = tuple[
    Integration,
    Info,
]


def case_info() -> Case:
    app_id = "test"
    integration = Integration(
        app_id=app_id,
        auth=BasicCredentials,
        exception_handlers=[],
        handle_errors=True,
    )

    @integration.register_capability(CapabilityName.LIST_ACCOUNTS)
    async def capability(request: Request[ListAccountsRequest]) -> Response[ListAccountsResponse]:
        return Response[ListAccountsResponse](
            response=ListAccountsResponse(
                accounts=[],
            ),
            raw_data=None,
        )

    expected_info = Info(
        app_id=app_id,
        capabilities=[CapabilityName.LIST_ACCOUNTS.value],
        authentication_schema={
            "properties": {
                "model": {
                    "const": "basic",
                    "default": "basic",
                    "enum": ["basic"],
                    "title": "Model",
                    "type": "string",
                },
                "username": {
                    "description": "Username",
                    "title": "Username",
                    "type": "string",
                },
                "password": {
                    "description": "Password",
                    "title": "Password",
                    "type": "string",
                },
            },
            "required": ["username", "password"],
            "title": "BasicCredentials",
            "type": "object",
        },
        capability_schema={
            CapabilityName.LIST_ACCOUNTS.value: CommandTypes(
                argument=json.dumps(
                    {
                        "$defs": {
                            "BasicCredentials": {
                                "properties": {
                                    "model": {
                                        "const": "basic",
                                        "default": "basic",
                                        "enum": ["basic"],
                                        "title": "Model",
                                        "type": "string",
                                    },
                                    "username": {
                                        "description": "Username",
                                        "title": "Username",
                                        "type": "string",
                                    },
                                    "password": {
                                        "description": "Password",
                                        "title": "Password",
                                        "type": "string",
                                    },
                                },
                                "required": ["username", "password"],
                                "title": "BasicCredentials",
                                "type": "object",
                            },
                            "ListAccountsRequest": {
                                "properties": {
                                    "custom_attributes": {
                                        "anyOf": [
                                            {"items": {"type": "string"}, "type": "array"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Custom Attributes",
                                    },
                                },
                                "title": "ListAccountsRequest",
                                "type": "object",
                            },
                            "OAuthCredentials": {
                                "properties": {
                                    "model": {
                                        "const": "oauth",
                                        "default": "oauth",
                                        "enum": ["oauth"],
                                        "title": "Model",
                                        "type": "string",
                                    },
                                    "access_token": {
                                        "description": "OAuth access token",
                                        "title": "Access Token",
                                        "type": "string",
                                    },
                                    "refresh_token": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "description": "OAuth refresh token",
                                        "title": "Refresh Token",
                                    },
                                    "scope": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "description": "OAuth scopes",
                                        "title": "Scope",
                                    },
                                },
                                "required": ["access_token"],
                                "title": "OAuthCredentials",
                                "type": "object",
                            },
                            "PaginationArgs": {
                                "properties": {
                                    "token": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Token",
                                    },
                                    "size": {
                                        "anyOf": [
                                            {"type": "integer"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Size",
                                    },
                                },
                                "title": "PaginationArgs",
                                "type": "object",
                            },
                        },
                        "properties": {
                            "request": {"$ref": "#/$defs/ListAccountsRequest"},
                            "auth": {
                                "discriminator": {
                                    "mapping": {
                                        "basic": "#/$defs/BasicCredentials",
                                        "oauth": "#/$defs/OAuthCredentials",
                                    },
                                    "propertyName": "model",
                                },
                                "oneOf": [
                                    {"$ref": "#/$defs/OAuthCredentials"},
                                    {"$ref": "#/$defs/BasicCredentials"},
                                ],
                                "title": "Auth",
                            },
                            "page": {
                                "anyOf": [
                                    {"$ref": "#/$defs/PaginationArgs"},
                                    {"type": "null"},
                                ],
                                "default": None,
                            },
                            "include_raw_data": {
                                "default": False,
                                "title": "Include Raw Data",
                                "type": "boolean",
                            },
                            "request_id": {
                                "anyOf": [
                                    {"type": "string"},
                                    {"type": "null"},
                                ],
                                "default": None,
                                "title": "Request Id",
                            },
                        },
                        "required": ["request", "auth"],
                        "title": "Request[ListAccountsRequest]",
                        "type": "object",
                    }
                ),
                output=json.dumps(
                    {
                        "$defs": {
                            "EncounteredErrorResponse": {
                                "properties": {
                                    "message": {
                                        "title": "Message",
                                        "type": "string",
                                    },
                                    "status_code": {
                                        "anyOf": [
                                            {"type": "integer"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Status Code",
                                    },
                                    "error_code": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Error Code",
                                    },
                                    "raised_by": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Raised By",
                                    },
                                    "raised_in": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Raised In",
                                    },
                                },
                                "required": ["message"],
                                "title": "EncounteredErrorResponse",
                                "type": "object",
                            },
                            "FoundAccountData": {
                                "properties": {
                                    "integration_specific_id": {
                                        "title": "Integration Specific Id",
                                        "type": "string",
                                    },
                                    "email": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Email",
                                    },
                                    "given_name": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Given Name",
                                    },
                                    "family_name": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Family Name",
                                    },
                                    "username": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Username",
                                    },
                                    "user_status": {
                                        "anyOf": [
                                            {"type": "string"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "User Status",
                                    },
                                    "extra_data": {
                                        "anyOf": [
                                            {"type": "object"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Extra Data",
                                    },
                                    "custom_attributes": {
                                        "anyOf": [
                                            {
                                                "additionalProperties": {"type": "string"},
                                                "type": "object",
                                            },
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Custom Attributes",
                                    },
                                },
                                "required": ["integration_specific_id"],
                                "title": "FoundAccountData",
                                "type": "object",
                            },
                            "ListAccountsResponse": {
                                "properties": {
                                    "accounts": {
                                        "items": {"$ref": "#/$defs/FoundAccountData"},
                                        "title": "Accounts",
                                        "type": "array",
                                    },
                                },
                                "required": ["accounts"],
                                "title": "ListAccountsResponse",
                                "type": "object",
                            },
                            "PaginationData": {
                                "additionalProperties": False,
                                "properties": {
                                    "token": {
                                        "title": "Token",
                                        "type": "string",
                                    },
                                    "size": {
                                        "anyOf": [
                                            {"type": "integer"},
                                            {"type": "null"},
                                        ],
                                        "default": None,
                                        "title": "Size",
                                    },
                                },
                                "required": ["token"],
                                "title": "PaginationData",
                                "type": "object",
                            },
                        },
                        "properties": {
                            "response": {"$ref": "#/$defs/ListAccountsResponse"},
                            "raw_data": {
                                "anyOf": [
                                    {"type": "object"},
                                    {"type": "null"},
                                ],
                                "default": None,
                                "title": "Raw Data",
                            },
                            "page": {
                                "anyOf": [
                                    {"$ref": "#/$defs/PaginationData"},
                                    {"type": "null"},
                                ],
                                "default": None,
                            },
                            "error": {
                                "anyOf": [
                                    {"$ref": "#/$defs/EncounteredErrorResponse"},
                                    {"type": "null"},
                                ],
                                "default": None,
                            },
                        },
                        "required": ["response"],
                        "title": "Response[ListAccountsResponse]",
                        "type": "object",
                    }
                ),
            ),
        },
    )
    return integration, expected_info



import typing as t
from enum import Enum

from pydantic import BaseModel, Field, Json


class EmptyModel(BaseModel):
    pass


class CommandTypes(BaseModel):
    argument: Json[t.Any] = Field(..., description="OpenAPI Component")
    output: Json[t.Any] = Field(..., description="OpenAPI Component")


class AppCategory(str, Enum):
    HR_AND_LEARNING = "HR_AND_LEARNING"
    OFFICE_AND_LEGAL = "OFFICE_AND_LEGAL"
    SALES_AND_SUPPORT = "SALES_AND_SUPPORT"
    COMMERCE_AND_MARKETPLACES = "COMMERCE_AND_MARKETPLACES"
    IT_AND_SECURITY = "IT_AND_SECURITY"
    COMMUNICATION = "COMMUNICATION"
    DESIGN_AND_CREATIVITY = "DESIGN_AND_CREATIVITY"
    OTHER = "OTHER"
    MARKETING_AND_ANALYTICS = "MARKETING_AND_ANALYTICS"
    DEVELOPERS = "DEVELOPERS"
    ACCOUNTING_AND_FINANCE = "ACCOUNTING_AND_FINANCE"
    COLLABORATION = "COLLABORATION"
    CONTENT_AND_SOCIAL_MEDIA = "CONTENT_AND_SOCIAL_MEDIA"
    INTERNAL = "INTERNAL"


class Info(BaseModel):
    app_id: str
    capabilities: list[str]
    authentication_schema: dict[str, t.Any] | None = None
    capability_schema: dict[str, CommandTypes]
    logo_url: str | None = None
    user_friendly_name: str | None = None
    description: str | None = None
    categories: list[AppCategory] | None = None



import pytest
import pytest_cases
from connector.oai.capability import CapabilityName, get_oauth
from {name}.serializers.request import (
    {pascal}ListEntitlementsRequest,
)
from connector.generated import (
    ListEntitlementsResponse,
)
from {name}.integration import integration

from tests.type_definitions import ClientContextManager, ResponseBodyMap

@pytest.mark.skip(reason="Function not implemented yet, remove after implementation of tested function.")
@pytest_cases.parametrize_with_cases(
    ["args", "response_body_map", "expected_response"],
    cases=[
        "tests.test_basic_object_read.test_list_entitlements_cases",
    ],
)
def test_list_entitlements(
    httpx_async_client: ClientContextManager,
    args: {pascal}ListEntitlementsRequest,
    response_body_map: ResponseBodyMap,
    expected_response: ListEntitlementsResponse,
) -> None:
    with httpx_async_client(get_oauth(args).access_token, response_body_map):
        response = integration.dispatch(CapabilityName.LIST_ENTITLEMENTS, args.model_dump_json())

    assert response == expected_response.model_dump_json()



"""Utility class to manage connector implementation.

:py:class:`Integration` provides a single point to register Integration
capabilities.

By instantiating the :py:class:`Integration` you simply create a basic
integration without any real implementation. To actually implement any
capability, you have to define (async) function outside the class and
register them to the integration instance by decorating the
implementation with ``@integration.register_capability(name)``.

Capability function has to:
    * accept only one argument
    * return scalar response

The :py:class:`Integration` is as much type-hinted as possible and also
does several checks to ensure that the implementation "is correct".
Incorrect implementation should raise an error during application start
(fail fast).

What is checked (at application start):
    * capability name is known (defined in ``CapabilityName`` enum)
    * the types of accepted argument and returned value matches the
    capability interface
"""

import asyncio
import inspect
import json
import logging
import typing as t
from dataclasses import dataclass

from connector.generated import (
    AppCategory,
    BasicCredential,
    CapabilityName,
    CapabilitySchema,
    Error,
    ErrorCode,
    ErrorResponse,
    Info,
    InfoResponse,
    OAuthCredential,
)
from connector.oai.capability import (
    CapabilityCallableProto,
    generate_capability_schema,
    get_capability_annotations,
    validate_capability,
)
from connector.oai.errors import ErrorMap, handle_exception
from pydantic import BaseModel

AuthSetting: t.TypeAlias = t.Union[
    t.Type[OAuthCredential],
    t.Type[BasicCredential],
]

logger = logging.getLogger(__name__)


class IntegrationError(Exception):
    """Base class for exceptions raised by Integration."""


class DuplicateCapabilityError(IntegrationError):
    """Raised when registering the same capability repeatedly."""

    def __init__(self, capability_name: CapabilityName) -> None:
        super().__init__(f"{capability_name} already registered")


class InvalidAppIdError(IntegrationError):
    """Raised when app_id is not valid.

    Most probably, empty or containing only whitespaces.
    """


@dataclass
class DescriptionData:
    logo_url: str | None = None
    user_friendly_name: str | None = None
    description: str | None = None
    categories: list[AppCategory] | None = None


class Integration:
    app_id: str

    def __init__(
        self,
        app_id: str,
        exception_handlers: ErrorMap,
        auth: AuthSetting,
        handle_errors: bool = True,
        description_data: DescriptionData | None = None,
    ):
        self.app_id = app_id.strip()
        self.auth = auth
        self.exception_handlers = exception_handlers
        self.handle_errors = handle_errors
        self.description_data = description_data or DescriptionData()

        if len(self.app_id) == 0:
            raise InvalidAppIdError

        self.capabilities: dict[CapabilityName, CapabilityCallableProto[t.Any]] = {}

    def register_capability(
        self,
        name: CapabilityName,
    ) -> t.Callable[
        [CapabilityCallableProto[t.Any]],
        CapabilityCallableProto[t.Any],
    ]:
        """Add implementation of specified capability.

        This function is expected to be used as a decorator for a
        capability implementation.

        Raises
        ------
        RuntimeError:
            When capability is registered more that once.
        """
        if name in self.capabilities:
            raise DuplicateCapabilityError(name)

        def decorator(
            func: CapabilityCallableProto[t.Any],
        ) -> CapabilityCallableProto[t.Any]:
            validate_capability(name, func)
            self.capabilities[name] = func
            return func

        return decorator

    def dispatch(self, name: CapabilityName, request_string: str) -> str:
        """Call implemented capability, returning the result.

        Raises
        ------
        NotImplementedError:
            When capability is not implemented (or registered)
        """
        try:
            capability = self.capabilities[name]
        except KeyError:
            if self.handle_errors:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message=f"Capability '{name.value}' is not implemented.",
                        error_code="connector.not_implemented",
                        std_error=ErrorCode.NOT_IMPLEMENTED,
                    ),
                ).model_dump_json()

            raise NotImplementedError from None

        request_annotation, _ = get_capability_annotations(capability)
        request = request_annotation(**json.loads(request_string))

        try:
            if inspect.iscoroutinefunction(capability):
                response = asyncio.run(
                    capability(request) if self.handle_errors else capability(request)
                )
            else:
                response = t.cast(
                    BaseModel,
                    (capability(request) if self.handle_errors else capability(request)),
                )

            return response.model_dump_json()
        except Exception as e:
            return handle_exception(
                self.exception_handlers, e, capability, self.app_id
            ).model_dump_json()

    def info(self) -> InfoResponse:
        """Provide information about implemented capabilities.

        Json schema describing implemented capabilities and their
        interface is returned. The authentication schema is also
        included.
        """
        capability_names = sorted(self.capabilities.keys())
        capability_schema: dict[str, CapabilitySchema] = {}
        for capability_name in capability_names:
            command_types = generate_capability_schema(
                capability_name, self.capabilities[capability_name]
            )
            capability_schema[capability_name] = CapabilitySchema(
                argument=command_types.argument,
                output=command_types.output,
            )
        return InfoResponse(
            response=Info(
                app_id=self.app_id,
                capabilities=capability_names,
                capability_schema=capability_schema,
                logo_url=self.description_data.logo_url,
                user_friendly_name=self.description_data.user_friendly_name,
                description=self.description_data.description,
                categories=self.description_data.categories,
            )
        )



from types import FunctionType
from typing import Any
from unittest import (
    TestCase,
)

from connector.errors import ConnectorError, ErrorCodes, ExceptionHandler
from connector.serializers.lumos import (
    EncounteredErrorResponse,
)
from connector.serializers.response import Response
from connector.sync_.exception_handler import connector_handler


class TestExceptionHandler(TestCase):
    def test_class_decorator_decorates_methods_correctly(self):
        # Mock a class with some methods
        class TestClass:
            def method1(self):
                pass

            def method2(self):
                pass

        # Call the class decorator
        decorated_class = connector_handler([])(TestClass)

        # Verify that methods are decorated
        self.assertTrue(hasattr(decorated_class.method1, "__wrapped__"))
        self.assertTrue(hasattr(decorated_class.method2, "__wrapped__"))

    def test_class_decorator_handles_exceptions(self):
        # Mock a class with some methods
        class TestClass:
            def method1():
                raise ValueError("Error 1")  # Raise a ValueError

        # Call the class decorator
        decorated_class = connector_handler([])(TestClass)
        raised = False
        response = None

        try:
            response = decorated_class.method1()
        except ValueError:
            raised = True

        # Verify that the exception was not raised and handled gracefully
        self.assertFalse(raised, "ValueError not raised")
        self.assertLogs("connector.errors", level="ERROR")
        self.assertIsInstance(response, Response)
        self.assertIsInstance(response.response, EncounteredErrorResponse)

    def test_class_decorator_handles_undefined_exceptions(self):
        # Mock a custom error class
        class CustomError(Exception):
            pass

        # Mock a class with some methods
        class TestClass:
            def method1():
                raise CustomError("Custom Error has been handled")  # Raise a custom error

        # Call the class decorator
        decorated_class = connector_handler([])(TestClass)
        raised = False
        response = None

        try:
            response = decorated_class.method1()
        except Exception:
            raised = True

        # Verify that the exception was not raised and handled gracefully
        self.assertFalse(raised, "CustomError not raised")
        self.assertLogs("connector.errors", level="ERROR")
        self.assertEqual(response.response.message, "Custom Error has been handled")

    def test_class_decorator_handles_connector_exception(self):
        # Mock a class with some methods
        class TestClass:
            app_id = "test"

            def method1():
                raise ConnectorError(
                    "ConnectorError has been raised.", ErrorCodes.INTERNAL_ERROR
                )  # Raise a ConnectorError

        # Call the class decorator
        decorated_class = connector_handler([])(TestClass)
        raised = False
        response = None

        try:
            response = decorated_class.method1()
        except ConnectorError:
            raised = True

        # Verify that the exception was not raised and handled gracefully, with an error code
        self.assertFalse(raised, "ConnectorError not raised")
        self.assertEqual(response.response.message, "ConnectorError has been raised.")
        self.assertEqual(response.response.error_code, "test.internal_error")

    def test_class_decorator_handles_custom_handler(self):
        # Mock a class with some methods
        class TestClass:
            app_id = "test"

            def method1():
                raise ValueError("ValueError has been raised.")  # Raise a ConnectorError

        # Create a custom handler class
        class CustomHandler(ExceptionHandler):
            @staticmethod
            def handle(
                e: Exception,
                original_func: FunctionType,
                response: Response[Any],
                error_code: str | ErrorCodes | None = None,
            ) -> Response[Any]:
                response.response.error_code = "custom_error_code"
                return response

        # Call the class decorator
        decorated_class = connector_handler([(ValueError, CustomHandler, None)])(TestClass)
        raised = False
        response = None

        try:
            response = decorated_class.method1()
        except ConnectorError:
            raised = True

        # Verify that our custom handler was ran and the exception was handled gracefully
        self.assertFalse(raised, "Error not raised")
        self.assertEqual(response.response.message, "ValueError has been raised.")
        self.assertEqual(response.response.error_code, "custom_error_code")



from datetime import datetime
from enum import Enum
from typing import Any, Dict, Generic, NewType, Optional, TypeAlias, TypeVar

from pydantic import BaseModel, ConfigDict, Field

NextPageTokenType = NewType("NextPageTokenType", str)

RawDataType: TypeAlias = dict[str, Any] | None


class AccountReference(BaseModel):
    """Account reference.

    Mostly used for arguments when we need to reference the account to
    which the operations is related.
    """

    integration_specific_id: str


class AccountType(str, Enum):
    SERVICE_ACCOUNT = "SERVICE_ACCOUNT"


class FoundAccountData(BaseModel):
    integration_specific_id: str
    email: Optional[str] = None
    given_name: Optional[str] = None
    family_name: Optional[str] = None
    username: Optional[str] = None
    user_status: Optional[str] = None
    extra_data: Optional[Dict[str, Any]] = None
    custom_attributes: Optional[Dict[str, str]] = None
    account_type: Optional[AccountType] = None


class CustomAttributeType(str, Enum):
    """Indicates the type of the attribute"""

    STRING = "STRING"
    USER = "USER"


class CustomAttributeCustomizedType(str, Enum):
    """Indicates the type of the entity that would own the attribue"""

    ACCOUNT = "ACCOUNT"
    ENTITLEMENMT = "ENTITLEMENT"
    RESOURCE = "RESOURCE"


class CustomAttributeSchema(BaseModel):
    customized_type: CustomAttributeCustomizedType
    name: str
    attribute_type: CustomAttributeType


class FoundResourceData(BaseModel):
    integration_specific_id: str
    label: str
    resource_type: str
    extra_data: Optional[Dict[str, Any]] = None


class FoundEntitlementData(BaseModel):
    integration_specific_id: str
    integration_specific_resource_id: str
    entitlement_type: str
    is_assignable: Optional[bool] = None
    label: Optional[str] = None
    extra_data: Optional[Dict[str, Any]] = None


class FoundEntitlementAssociation(BaseModel):
    integration_specific_entitlement_id: str
    account: FoundAccountData
    integration_specific_resource_id: str


class EncounteredErrorResponse(BaseModel):
    message: str
    status_code: Optional[int] = None
    error_code: Optional[str] = None
    raised_by: Optional[str] = None
    raised_in: Optional[str] = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


Response = TypeVar("Response")


class PageCursor(BaseModel):
    next: NextPageTokenType | None


class PaginationPageArgs(BaseModel):
    """Schema of request page parameters."""

    model_config = ConfigDict(extra="forbid")

    token: NextPageTokenType | None = Field(
        default=None,
        description=(
            "Next page token returned from previous method call in ``page.token``. The value of the next page token is "
            "opaque to the caller and should not be modified by any means."
        ),
    )
    size: int | None = Field(
        default=None,
        gt=0,
        description="Prefered size of result page. Connector can decrease the requested size if needed.",
    )


class PaginationPageResp(BaseModel):
    """Schema of request page parameters."""

    model_config = ConfigDict(extra="forbid")

    token: NextPageTokenType = Field(
        default=None,
        description=(
            "Next page token returned from previous method call in ``page.token``. The value of the next page token is "
            "opaque to the caller and should not be modified by any means."
        ),
    )
    size: int | None = Field(
        default=None,
        gt=0,
        description="Prefered size of result page. Connector can decrease the requested size if needed.",
    )


class PaginationArgs(BaseModel):
    """Definition of pagination schema for requests.

    To add pagination support to your request schema, simply inherit
    from this class, e.g.: ::

        class ListUserArgs(PaginationArgs, DefaultListUserArgs):
            pass
    """

    page: PaginationPageArgs | None = None

    @property
    def pagination_token(self) -> NextPageTokenType | None:
        if self.page is None:
            return None

        return self.page.token

    @property
    def pagination_size(self) -> int | None:
        if self.page is None:
            return None

        return self.page.size


class ResponseWrapper(BaseModel, Generic[Response]):
    response: Response
    raw_data: RawDataType = None
    cursor: PageCursor | None = Field(
        default=None,
        deprecated="cursor attribute is deprecated, use `page` instead",
    )
    page: PaginationPageResp | None = None


class BaseArgs(BaseModel):
    include_raw_data: bool = Field(default=False)


class ListAccountsArgsBase(BaseArgs):
    custom_attributes: Optional[list[str]] = Field(None)


class ListCustomAttributesSchemaArgsBase(BaseArgs):
    pass


class ListCustomAttributesSchemaResp(ResponseWrapper[list[CustomAttributeSchema]]):
    pass


class ListAccountsResp(ResponseWrapper[list[FoundAccountData]]):
    pass


class ValidateCredentialsArgsBase(BaseArgs):
    pass


class ValidateCredentialsResp(ResponseWrapper[bool]):
    pass


class GetAccountArgsBase(BaseArgs):
    pass


class GetAccountResp(ResponseWrapper[FoundAccountData]):
    pass


class ListResourcesArgsBase(BaseArgs):
    resource_type: str


class ListResourcesResp(ResponseWrapper[list[FoundResourceData]]):
    pass


class GetResourceArgsBase(BaseArgs):
    resource_type: str
    integration_specific_id: str


class GetResourceResp(ResponseWrapper[FoundResourceData]):
    pass


class ListEntitlementsArgsBase(BaseArgs):
    resource_type: str
    resource_integration_specific_id: str


class ListEntitlementsResp(ResponseWrapper[list[FoundEntitlementData]]):
    pass


class FindEntitlementAssociationsArgsBase(BaseArgs, PaginationArgs):
    pass


class FindEntitlementAssociationsResp(ResponseWrapper[list[FoundEntitlementAssociation]]):
    pass


class AssignEntitlementArgsBase(BaseArgs):
    account: FoundAccountData
    entitlement: FoundEntitlementData


class AssignEntitlementResp(ResponseWrapper[bool]):
    pass


class UnassignEntitlementArgsBase(BaseArgs):
    account: FoundAccountData
    entitlement: FoundEntitlementData


class UnassignEntitlementResp(ResponseWrapper[bool]):
    pass


class CreateAccount(BaseModel):
    email: str | None = None
    username: str | None = None
    given_name: str | None = None
    family_name: str | None = None
    user_status: str | None = None
    extra_data: Dict[str, Any] | None = None


class CreateAccountArgsBase(BaseArgs):
    account: CreateAccount
    entitlements: list[FoundEntitlementData]


class CreateAccountResp(ResponseWrapper[bool]):
    pass


class DeleteAccountArgsBase(BaseArgs):
    account_id: str


class DeleteAccountResp(ResponseWrapper[bool]):
    pass


class ActivateAccountArgsBase(BaseArgs):
    account_id: str


class ActivateAccountResp(ResponseWrapper[bool]):
    pass


class DeactivateAccountArgsBase(BaseArgs):
    account_id: str


class DeactivateAccountResp(ResponseWrapper[bool]):
    pass


class LastActivityData(BaseModel):
    account_id: str
    event_type: str
    happened_at: datetime


class GetLastActivityArgsBase(BaseArgs):
    account_ids: list[str]


class GetLastActivityResp(ResponseWrapper[list[LastActivityData]]):
    pass


class ErrorResp(ResponseWrapper[EncounteredErrorResponse]):
    error: bool = True
    pass



"""Cases for testing ``delete_account`` operation."""

import typing as t

import httpx
from connector.generated import (
    DeleteAccountResponse,
    DeletedAccount,
    Error,
    ErrorCode,
    ErrorResponse,
)
from connector.utils.test import http_error_message

from {name}.integration import BASE_URL
from {name}.serializers.request import (
    {pascal}DeleteAccount,
    {pascal}DeleteAccountRequest,
)
from tests.conftest import INVALID_AUTH, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    {pascal}DeleteAccountRequest,
    ResponseBodyMap,
    DeleteAccountResponse | ErrorResponse,
]


def case_delete_account_204() -> Case:
    """Successful deletion request."""
    args = {pascal}DeleteAccountRequest(
        request={pascal}DeleteAccount(
            account_id="1",
        ),
        auth=VALID_AUTH,
        include_raw_data=True,
    )
    response_body_map = {{
        "DELETE": {{
            f"/users/{{args.request.account_id}}": MockedResponse(
                status_code=httpx.codes.NO_CONTENT,
                response_body=None,
            ),
        }},
    }}
    expected_response = DeleteAccountResponse(
        response=DeletedAccount(deleted=True),
        raw_data={{
            f"{{BASE_URL}}/users/{{args.request.account_id}}": None,
        }},
    )
    return args, response_body_map, expected_response


def case_delete_account_404() -> Case:
    """Not found request should fail."""
    args = {pascal}DeleteAccountRequest(
        request={pascal}DeleteAccount(
            account_id="non_existent",
        ),
        auth=VALID_AUTH,
    )
    response_body_map = {{
        "DELETE": {{
            f"/users/{{args.request.account_id}}": MockedResponse(
                status_code=httpx.codes.NOT_FOUND,
                response_body={{
                    "error": {{
                        "message": "Not found",
                        "code": 2100,
                    }},
                }},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                f"{{BASE_URL}}/users/{{args.request.account_id}}",
                httpx.codes.NOT_FOUND,
            ),
            status_code=httpx.codes.NOT_FOUND,
            error_code=ErrorCode.NOT_FOUND,
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:delete_account",
        ),
    )
    return args, response_body_map, expected_response


def case_delete_account_401() -> Case:
    """Unauthorized request should fail."""
    args = {pascal}DeleteAccountRequest(
        request={pascal}DeleteAccount(
            account_id="1",
        ),
        auth=INVALID_AUTH,
    )
    response_body_map = {{
        "DELETE": {{
            f"/users/{{args.request.account_id}}": MockedResponse(
                status_code=httpx.codes.UNAUTHORIZED,
                response_body=None,
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                f"{{BASE_URL}}/users/{{args.request.account_id}}",
                httpx.codes.UNAUTHORIZED,
            ),
            status_code=httpx.codes.UNAUTHORIZED,
            error_code=ErrorCode.UNAUTHORIZED,
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:delete_account",
        ),
    )
    return args, response_body_map, expected_response



from typing import TypeAlias

import httpx
from connector.oai.capability import CapabilityName, get_oauth, get_page
from connector.oai.errors import HTTPHandler
from connector.oai.integration import Integration
from connector.generated import (
    ActivateAccountResponse,
    ActivatedAccount,
    AssignEntitlementResponse,
    AssignedEntitlement,
    CreateAccountResponse,
    CreatedAccount,
    DeactivateAccountResponse,
    DeactivatedAccount,
    DeleteAccountResponse,
    DeletedAccount,
    FindEntitlementAssociationsResponse,
    FoundAccountData,
    FoundEntitlementAssociation,
    GetLastActivityResponse,
    LastActivityData,
    ListAccountsResponse,
    ListCustomAttributesSchemaResponse,
    ListEntitlementsResponse,
    ListResourcesResponse,
    OAuthCredential,
    Page,
    UnassignEntitlementResponse,
    UnassignedEntitlement,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)
from {name}.auth import Auth
from {name}.serializers.pagination import DEFAULT_PAGE_SIZE, NextPageToken, Pagination
from {name}.serializers.request import (
    {pascal}ActivateAccount,
    {pascal}ActivateAccountRequest,
    {pascal}AssignEntitlement,
    {pascal}AssignEntitlementRequest,
    {pascal}CreateAccount,
    {pascal}CreateAccountRequest,
    {pascal}DeactivateAccount,
    {pascal}DeactivateAccountRequest,
    {pascal}DeleteAccount,
    {pascal}DeleteAccountRequest,
    {pascal}FindEntitlementAssociations,
    {pascal}FindEntitlementAssociationsRequest,
    {pascal}GetLastActivity,
    {pascal}GetLastActivityRequest,
    {pascal}ListAccounts,
    {pascal}ListAccountsRequest,
    {pascal}ListCustomAttributesSchema,
    {pascal}ListCustomAttributesSchemaRequest,
    {pascal}ListEntitlements,
    {pascal}ListEntitlementsRequest,
    {pascal}ListResources,
    {pascal}ListResourcesRequest,
    {pascal}UnassignEntitlement,
    {pascal}UnassignEntitlementRequest,
    {pascal}ValidateCredentials,
    {pascal}ValidateCredentialsRequest,
)

BASE_URL = "https://scaffold.com"

RequestData: TypeAlias = (
    {pascal}ActivateAccountRequest |
    {pascal}AssignEntitlementRequest |
    {pascal}CreateAccountRequest |
    {pascal}DeactivateAccountRequest |
    {pascal}DeleteAccountRequest |
    {pascal}FindEntitlementAssociationsRequest |
    {pascal}GetLastActivityRequest |
    {pascal}ListAccountsRequest |
    {pascal}ListCustomAttributesSchemaRequest |
    {pascal}ListEntitlementsRequest |
    {pascal}ListResourcesRequest |
    {pascal}UnassignEntitlementRequest |
    {pascal}ValidateCredentialsRequest
)

def build_client(request: RequestData) -> httpx.AsyncClient:
    """Prepare client context manager for calling {title} API."""
    return httpx.AsyncClient(
        auth=Auth(access_token=get_oauth(request).access_token),
        base_url=BASE_URL,
    )


integration = Integration(
    app_id="{hyphenated_name}",
    auth=OAuthCredential,
    exception_handlers=[
        (httpx.HTTPStatusError, HTTPHandler, None),
    ],
)

@integration.register_capability(CapabilityName.VALIDATE_CREDENTIALS)
async def validate_credentials(
    args: {pascal}ValidateCredentialsRequest,
) -> ValidateCredentialsResponse:
    async with build_client(args) as client:
        r = await client.get("/users", params={{"limit": 1}})
        r.raise_for_status()
        data = r.json()

    return ValidateCredentialsResponse(
        response=ValidatedCredentials(valid=True),
        raw_data={{
            f"{{BASE_URL}}/users?limit=1": data,
        }}
        if args.include_raw_data
        else None,
    )

@integration.register_capability(CapabilityName.LIST_ACCOUNTS)
async def list_accounts(args: {pascal}ListAccountsRequest) -> ListAccountsResponse:
    endpoint = "/users"
    try:
        current_pagination = NextPageToken(get_page(args).token).paginations()[0]
    except IndexError:
        current_pagination = Pagination.default(endpoint)

    page_size = get_page(args).size or DEFAULT_PAGE_SIZE
    async with build_client(args) as client:
        r = await client.get(
            endpoint,
            params={{"limit": page_size, "offset": current_pagination.offset}},
        )
        r.raise_for_status()
        data = r.json()
        accounts: list[FoundAccountData] = []

        if True:
            next_pagination = [
                Pagination(
                    endpoint=endpoint,
                    offset=current_pagination.offset + len(accounts),
                )
            ]
        else:
            next_pagination = []

        next_page_token = NextPageToken.from_paginations(next_pagination).token

    return ListAccountsResponse(
        response=accounts,
        raw_data = {{
            f"{{BASE_URL}}/users?limit=1": data,
        }} if args.include_raw_data else None,
        page=Page(
            token=next_page_token,
            size=page_size,
        )
        if next_page_token
        else None,
    )

# @integration.register_capability(CapabilityName.LIST_RESOURCES)
async def list_resources(args: {pascal}ListResourcesRequest) -> ListResourcesResponse:
    raise NotImplementedError

# @integration.register_capability(CapabilityName.LIST_ENTITLEMENTS)
async def list_entitlements(
    args: {pascal}ListEntitlementsRequest,
) -> ListEntitlementsResponse:
    raise NotImplementedError

# @integration.register_capability(CapabilityName.FIND_ENTITLEMENT_ASSOCIATIONS)
async def find_entitlement_associations(
    args: {pascal}FindEntitlementAssociationsRequest,
) -> FindEntitlementAssociationsResponse:
    raise NotImplementedError

# @integration.register_capability(CapabilityName.ASSIGN_ENTITLEMENT)
async def assign_entitlement(args: {pascal}AssignEntitlementRequest) -> AssignEntitlementResponse:
    raise NotImplementedError

# @integration.register_capability(CapabilityName.UNASSIGN_ENTITLEMENT)
async def unassign_entitlement(args: {pascal}UnassignEntitlementRequest) -> UnassignEntitlementResponse:
    raise NotImplementedError

# @integration.register_capability(CapabilityName.LIST_CUSTOM_ATTRIBUTES_SCHEMA)
async def list_custom_attributes_schema(args: {pascal}ListCustomAttributesSchemaRequest) -> ListCustomAttributesSchemaResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.CREATE_ACCOUNT)
async def create_account(
    args: {pascal}CreateAccountRequest,
) -> CreateAccountResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.DELETE_ACCOUNT)
async def delete_account(
    args: {pascal}DeleteAccountRequest,
) -> DeleteAccountResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.ACTIVATE_ACCOUNT)
async def activate_account(
    args: {pascal}ActivateAccountRequest,
) -> ActivateAccountResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.DEACTIVATE_ACCOUNT)
async def deactivate_account(
    args: {pascal}DeactivateAccountRequest,
) -> DeactivateAccountResponse:
    raise NotImplementedError



"""Cases for testing ``list_accounts`` operation."""

import typing as t

import httpx
from connector.generated import (
    ListAccountsResponse,
    Error,
    ErrorCode,
    ErrorResponse,
    Page,
)
from connector.utils.test import http_error_message

from {name}.integration import BASE_URL
from {name}.serializers.request import (
    {pascal}ListAccounts,
    {pascal}ListAccountsRequest,
)
from tests.conftest import INVALID_AUTH, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    {pascal}ListAccountsRequest,
    ResponseBodyMap,
    ListAccountsResponse | ErrorResponse,
]


def case_list_accounts_200() -> Case:
    """Successful request."""
    args = {pascal}ListAccountsRequest(
        request={pascal}ListAccounts(),
        auth=VALID_AUTH,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=5&offset=0": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ListAccountsResponse(
        response=[],
        raw_data=None,
        page=Page(
            token="9182a8656e64706f696e74a62f7573657273a66f666673657400",
            size=5,
        ),
    )
    return args, response_body_map, expected_response


def case_list_accounts_200_no_accounts() -> Case:
    """No accounts found."""
    args = {pascal}ListAccountsRequest(
        request={pascal}ListAccounts(),
        auth=VALID_AUTH,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=5&offset=0": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ListAccountsResponse(
        response=[],
        raw_data=None,
        page=Page(
            token="9182a8656e64706f696e74a62f7573657273a66f666673657400",
            size=5,
        ),
    )
    return args, response_body_map, expected_response


def case_list_accounts_401() -> Case:
    """Unauthorized request should fail."""
    args = {pascal}ListAccountsRequest(
        request={pascal}ListAccounts(),
        auth=INVALID_AUTH,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=5&offset=0": MockedResponse(
                status_code=httpx.codes.UNAUTHORIZED,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                f"{{BASE_URL}}/users?limit=5&offset=0",
                httpx.codes.UNAUTHORIZED,
            ),
            status_code=httpx.codes.UNAUTHORIZED,
            error_code=ErrorCode.UNAUTHORIZED,
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:list_accounts",
        ),
    )
    return args, response_body_map, expected_response



"""Test cases for ``Integration.dispatch`` function.

Those cases should lead to raised error.
"""

import json

from connector.capability import CapabilityName
from connector.integration import Integration
from connector.serializers.request import (
    BasicCredentials,
    ListAccountsRequest,
    Request,
)
from connector.serializers.response import (
    ListAccountsResponse,
    Response,
)

Case = tuple[
    Integration,
    CapabilityName,
    str,
    type[Exception],
]


def case_dispatch_not_implemented_handled() -> Case:
    """Unhandled call to unimplemented capability should raise."""
    integration = Integration(
        app_id="test",
        auth=BasicCredentials,
        exception_handlers=[],
        handle_errors=False,
    )
    capability_name = CapabilityName.LIST_ACCOUNTS
    # don't have to care about actual request data, integration should reject the call before it touches it
    request = "{}"
    exception_type = NotImplementedError
    return integration, capability_name, request, exception_type


def case_dispatch_async_not_handled_error() -> Case:
    """Unhandled error in async capability should raise."""
    integration = Integration(
        app_id="test",
        auth=BasicCredentials,
        exception_handlers=[],
        handle_errors=False,
    )

    class CustomException(Exception):
        pass

    @integration.register_capability(CapabilityName.LIST_ACCOUNTS)
    async def list_accounts(req: Request[ListAccountsRequest]) -> Response[ListAccountsResponse]:
        raise CustomException

    capability_name = CapabilityName.LIST_ACCOUNTS
    request = json.dumps(
        {
            "auth": {"model": "basic", "username": "user", "password": "pass"},
            "request": {},
        }
    )
    exception_type = CustomException

    return integration, capability_name, request, exception_type


def case_dispatch_sync_not_handled_error() -> Case:
    """Unhandled error in sync capability should raise."""
    integration = Integration(
        app_id="test",
        auth=BasicCredentials,
        exception_handlers=[],
        handle_errors=False,
    )

    class CustomException(Exception):
        pass

    @integration.register_capability(CapabilityName.LIST_ACCOUNTS)
    def list_accounts(req: Request[ListAccountsRequest]) -> Response[ListAccountsResponse]:
        raise CustomException

    capability_name = CapabilityName.LIST_ACCOUNTS
    request = json.dumps(
        {
            "auth": {"model": "basic", "username": "user", "password": "pass"},
            "request": {},
        }
    )
    exception_type = CustomException

    return integration, capability_name, request, exception_type
